"""
PC-BASIC - display.graphics
Graphics operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import math
import operator

from itertools import islice

from ...compat import int2byte
from ..base import error
from ..base import tokens as tk
from ..base import bytematrix
from .. import values
from .. import mlparser


ZERO_TILE = bytematrix.ByteMatrix(1, 8)


class GraphicsViewPort(object):
    """Graphics viewport (clip area) functions."""

    def __init__(self, pixel_buffer):
        """Initialise graphics viewport."""
        self._pixels = pixel_buffer
        self._max_width, self._max_height = self._pixels.width, self._pixels.height
        self._absolute = False
        self._rect = 0, 0, self._max_width-1, self._max_height-1
        self._active = False

    def unset(self):
        """Unset the graphics viewport."""
        self._absolute = False
        self._rect = 0, 0, self._max_width-1, self._max_height-1
        self._active = False

    def set(self, x0, y0, x1, y1, absolute):
        """Set the graphics viewport."""
        # VIEW orders the coordinates
        x0, x1 = min(x0, x1), max(x0, x1)
        y0, y1 = min(y0, y1), max(y0, y1)
        self._absolute = absolute
        self._rect = x0, y0, x1, y1
        self._active = True

    @property
    def active(self):
        """Return whether the graphics viewport is set."""
        return self._active

    def set_page(self, pixel_buffer):
        """Set the pixel buffer (without adjusting the viewport."""
        assert pixel_buffer.width == self.width
        assert pixel_buffer.height == self.height
        self._pixels = pixel_buffer

    @property
    def height(self):
        """Height of the viewport."""
        return self._rect[3] - self._rect[1] + 1

    @property
    def width(self):
        """Width of the viewport."""
        return self._rect[2] - self._rect[0] + 1

    def __setitem__(self, index, data):
        """Set pixels in viewport."""
        self._pixels[self._convert_slice(index)] = data

    def __getitem__(self, index):
        """Get pixels in viewport."""
        yslice, xslice = index
        if not isinstance(yslice, slice) and not isinstance(xslice, slice):
            # single pixel read can go outside of viewport
            x, y = self._convert_coords(xslice, yslice)
            return self._pixels[y, x]
        return self._pixels[self._convert_slice(index)]

    def get_bounds(self):
        """Return the graphics viewport bounds, in viewport coordinates."""
        if self._absolute:
            return self._rect
        return 0, 0, self.width-1, self.height-1

    def contains(self, x, y):
        """Return whether the specified point is within the graphics view (boundaries inclusive)."""
        vx0, vy0, vx1, vy1 = self.get_bounds()
        return vx0 <= x <= vx1 and vy0 <= y <= vy1

    def get_mid(self):
        """Get the midpoint of the current graphics view, in viewpoint coordinates."""
        x0, y0, x1, y1 = self.get_bounds()
        # +1 to match GW-BASIC
        x, y = (x1-x0) // 2 + 1, (y1-y0) // 2 + 1
        if self._absolute:
            return x0 + x, y0 + y
        return x, y

    def cutoff_coord(self, x, y):
        """Ensure coordinates are within screen + 1 pixel."""
        abs_x, abs_y = self._convert_coords(x, y)
        offs_x, offs_y = x - abs_x, y - abs_y
        # clip absolute coornates to physical screen bounds plus one pixel
        abs_x = min(self._max_width, max(-1, abs_x))
        abs_y = min(self._max_height, max(-1, abs_y))
        # return viewpoint coordinates
        return abs_x + offs_x, abs_y + offs_y

    def _convert_coords(self, x, y):
        """Retrieve absolute coordinates for viewport coordinates."""
        if self._absolute:
            return x, y
        else:
            return x + self._rect[0], y + self._rect[1]

    def _convert_slice(self, slice_tuple):
        """Convert viewport to absolute slice tuple."""
        yslice, xslice = slice_tuple
        xmin, ymin, xmax, ymax = self.get_bounds()
        if not isinstance(yslice, slice) and not isinstance(xslice, slice):
            # single pixel
            if not self.contains(xslice, yslice):
                return slice(0, 0), slice(0, 0)
            xslice, yslice = self._convert_coords(xslice, yslice)
            return yslice, xslice
        # note that integer indices are converted to slices n:n+1
        if not isinstance(yslice, slice):
            yslice = slice(yslice, yslice+1)
        if not isinstance(xslice, slice):
            xslice = slice(xslice, xslice+1)
        assert yslice.step is None
        assert xslice.step is None
        y0, y1 = yslice.start, yslice.stop
        x0, x1 = xslice.start, xslice.stop
        if x0 is None:
            x0 = xmin
        if y0 is None:
            y0 = ymin
        if x1 is None:
            x1 = xmax
        if y1 is None:
            y1 = ymax
        # clip to bounds
        x0 = max(x0, xmin)
        y0 = max(y0, ymin)
        x1 = min(x1, xmax+1)
        y1 = min(y1, ymax+1)
        # convert top-left and bottom-right coordinates
        x0, y0 = self._convert_coords(x0, y0)
        x1, y1 = self._convert_coords(x1, y1)
        # rebuild slices
        yslice = slice(y0, y1)
        xslice = slice(x0, x1)
        return yslice, xslice


class Graphics(object):
    """Graphics operations."""

    def __init__(self, input_methods, values, memory, aspect, colourmap):
        """Initialise graphics object."""
        # for apagenum and attr
        self._values = values
        self._memory = memory
        # for wait() in paint_
        self._input_methods = input_methods
        # memebers set on mode switch
        self._mode = None
        self._pages = None
        self._apage = None
        self.graph_view = None
        self._apagenum = None
        # last accessed coordinate, viewpoint-relative
        self._last_point = None
        # when WINDOW is active, DRAW pointer gets reset by other commands but not vice versa
        self._draw_current = None
        self._last_attr = None
        self._draw_scale = None
        self._draw_angle = None
        # screen aspect ratio: used to determine pixel aspect ratio, which is used by CIRCLE
        self._screen_aspect = aspect

    def init_mode(self, mode, pages, colourmap):
        """Initialise for new graphics mode."""
        self._mode = mode
        self._pages = pages
        self._colourmap = colourmap
        self._num_attr = colourmap.num_attr
        # set graphics viewport
        self.graph_view = GraphicsViewPort(self._pages[0].pixels)
        self._unset_window()
        self.reset()

    def reset(self):
        """Reset graphics state."""
        if self._mode.is_text_mode:
            return
        self._last_point = self.graph_view.get_mid()
        self._draw_current = None
        self._last_attr = self._mode.attr
        self._draw_scale = 4
        self._draw_angle = 0

    def set_attr(self, attr):
        """Set the current attribute."""
        self._attr = attr

    def set_page(self, apagenum):
        """Set the active page."""
        self._apagenum = apagenum
        self._apage = self._pages[apagenum]
        self.graph_view.set_page(self._apage.pixels)

    ### attributes

    def _get_attr_index(self, attr_index):
        """Get the colour attribute for the specified index."""
        if attr_index == -1:
            # foreground; graphics 'background' attrib is always 0
            attr, _, _, _ = self._colourmap.split_attr(self._attr)
            return attr
        elif not attr_index:
            return 0
        return min(self._num_attr-1, max(0, attr_index))

    ## VIEW graphics viewport

    def view_(self, args):
        """VIEW: Set/unset the graphics viewport and optionally draw a box."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        # VIEW SCREEN
        absolute = next(args)
        bounds = [
            values.to_int(_arg)
            for _arg in islice(args, 4)
        ]
        if not bounds:
            # VIEW SCREEN is a syntax error; just VIEW is OK
            error.throw_if(absolute, error.STX)
            return self._unset_view()
        x0, y0, x1, y1 = bounds
        error.range_check(0, self._mode.pixel_width-1, x0, x1)
        error.range_check(0, self._mode.pixel_height-1, y0, y1)
        error.throw_if(x0==x1 or y0 == y1)
        fill = next(args)
        if fill is not None:
            fill = values.to_int(fill)
        border = next(args)
        if border is not None:
            border = values.to_int(border)
        error.range_check(0, 255, fill)
        error.range_check(0, 255, border)
        list(args)
        self._set_view(x0, y0, x1, y1, absolute, fill, border)

    def _set_view(self, x0, y0, x1, y1, absolute, fill, border):
        """Set the graphics viewport and optionally draw a box (VIEW)."""
        # first unset the viewport so that we can draw the box
        fill = self._get_attr_index(fill)
        border = self._get_attr_index(border)
        self.graph_view.unset()
        if fill is not None:
            self._draw_box_filled(x0, y0, x1, y1, fill)
            self._last_attr = fill
        if border is not None:
            self._draw_box(x0-1, y0-1, x1+1, y1+1, border)
            self._last_attr = border
        self.graph_view.set(x0, y0, x1, y1, absolute)
        self._last_point = self.graph_view.get_mid()
        self._draw_current = None
        if self._window_bounds is not None:
            self._set_window(*self._window_bounds)

    def _unset_view(self):
        """Unset the graphics viewport."""
        self.graph_view.unset()
        self._last_point = self.graph_view.get_mid()
        self._draw_current = None
        if self._window_bounds is not None:
            self._set_window(*self._window_bounds)

    ### WINDOW logical coords

    def window_(self, args):
        """WINDOW: Set/unset the logical coordinate window."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        cartesian = not next(args)
        try:
            coords = [values.to_single(_arg).to_value() for _arg in islice(args, 4)]
        except StopIteration:
            coords = []
        if not coords:
            self._unset_window()
        else:
            x0, y0, x1, y1 = coords
            if x0 == x1 or y0 == y1:
                raise error.BASICError(error.IFC)
            list(args)
            self._set_window(x0, y0, x1, y1, cartesian)

    def _set_window(self, fx0, fy0, fx1, fy1, cartesian=True):
        """Set the logical coordinate window (WINDOW)."""
        if fy0 > fy1:
            fy0, fy1 = fy1, fy0
        if fx0 > fx1:
            fx0, fx1 = fx1, fx0
        if cartesian:
            fy0, fy1 = fy1, fy0
        x0, y0 = 0., 0.
        x1, y1 = self.graph_view.width-1, self.graph_view.height-1
        scalex = (x1-x0) / (fx1-fx0)
        scaley = (y1-y0) / (fy1-fy0)
        offsetx = x0 - fx0*scalex
        offsety = y0 - fy0*scaley
        self._window = scalex, scaley, offsetx, offsety
        self._window_bounds = fx0, fy0, fx1, fy1, cartesian

    def _unset_window(self):
        """Unset the logical coordinate window."""
        self._window = None
        self._window_bounds = None

    def _get_window_physical(self, fx, fy, step=False):
        """Convert logical to physical coordinates."""
        if self._window:
            scalex, scaley, offsetx, offsety = self._window
            if step:
                fx0, fy0 = self._get_window_logical(*self._last_point)
            else:
                fx0, fy0 = 0., 0.
            x = int(round(offsetx + (fx0+fx) * scalex))
            y = int(round(offsety + (fy0+fy) * scaley))
        else:
            x, y = self._last_point if step else (0, 0)
            x += int(round(fx))
            y += int(round(fy))
        # overflow check
        if x < -0x8000 or y < -0x8000 or x > 0x7fff or y > 0x7fff:
            raise error.BASICError(error.OVERFLOW)
        return x, y

    def _get_window_logical(self, x, y):
        """Convert physical to logical coordinates."""
        x, y = float(x), float(y)
        if self._window:
            scalex, scaley, offsetx, offsety = self._window
            return (x - offsetx) / scalex,  (y - offsety) / scaley
        else:
            return x, y

    def _get_window_scale(self, fx, fy):
        """Get logical to physical scale factor."""
        if self._window:
            scalex, scaley, _, _ = self._window
            x, y = int(round(fx * scalex)), int(round(fy * scaley))
        else:
            x, y = int(round(fx)), int(round(fy))
        error.range_check_err(-32768, 32767, x, error.OVERFLOW)
        error.range_check_err(-32768, 32767, y, error.OVERFLOW)
        return x, y

    ### PSET, POINT

    def pset_(self, args):
        """PSET: set a pixel to a given attribute, or foreground."""
        self._pset_preset(args, -1)

    def preset_(self, args):
        """PRESET: set a pixel to a given attribute, or background."""
        self._pset_preset(args, 0)

    def _pset_preset(self, args, default):
        """Set a pixel to a given attribute."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        step = next(args)
        x, y = (values.to_single(_arg).to_value() for _arg in islice(args, 2))
        attr_index = next(args)
        if attr_index is None:
            attr_index = default
        else:
            attr_index = values.to_int(attr_index)
            error.range_check(0, 255, attr_index)
        list(args)
        x, y = self._get_window_physical(x, y, step)
        attr = self._get_attr_index(attr_index)
        # record viewpoint-relative physical coordinates
        self._last_point = x, y
        self._draw_current = None
        self._last_attr = attr
        self.graph_view[y, x] = attr

    ### LINE

    def line_(self, args):
        """LINE: Draw a patterned line or box."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        step0 = next(args)
        x0, y0 = (
            None if arg is None else values.to_single(arg).to_value()
            for _, arg in zip(range(2), args)
        )
        step1 = next(args)
        x1, y1 = (values.to_single(_arg).to_value() for _arg in islice(args, 2))
        coord0 = x0, y0, step0
        coord1 = x1, y1, step1
        attr_index = next(args)
        if attr_index:
            attr_index = values.to_int(attr_index)
            error.range_check(0, 255, attr_index)
        shape, pattern = args
        if attr_index is None:
            attr_index = -1
        if pattern is None:
            pattern = 0xffff
        else:
            pattern = values.to_int(pattern)
        if coord0 != (None, None, None):
            x0, y0 = self._get_window_physical(*coord0)
        else:
            x0, y0 = self._last_point
        self._last_point = x0, y0
        x1, y1 = self._get_window_physical(*coord1)
        attr = self._get_attr_index(attr_index)
        if not shape:
            self._draw_line(x0, y0, x1, y1, attr, pattern)
        elif shape == b'B':
            self._draw_box(x0, y0, x1, y1, attr, pattern)
        elif shape == b'BF':
            self._draw_box_filled(x0, y0, x1, y1, attr)
        self._last_point = x1, y1
        self._draw_current = None
        self._last_attr = attr

    def _draw_line(self, x0, y0, x1, y1, attr, pattern=0xffff):
        """Draw a line between the given physical points."""
        # cut off any out-of-bound coordinates
        x0, y0 = self.graph_view.cutoff_coord(x0, y0)
        x1, y1 = self.graph_view.cutoff_coord(x1, y1)
        if y1 <= y0:
            # work from top to bottom, or from x1,y1 if at the same height. this matters for mask.
            x1, y1, x0, y0 = x0, y0, x1, y1
        # Bresenham algorithm
        dx, dy = abs(x1-x0), abs(y1-y0)
        steep = dy > dx
        if steep:
            x0, y0, x1, y1 = y0, x0, y1, x1
            dx, dy = dy, dx
        sx = 1 if x1 > x0 else -1
        sy = 1 if y1 > y0 else -1
        mask = 0x8000
        line_error = dx // 2
        x, y = x0, y0
        for x in range(x0, x1+sx, sx):
            if pattern & mask != 0:
                if steep:
                    # set point (y, x)
                    self.graph_view[x, y] = attr
                else:
                    self.graph_view[y, x] = attr
            mask >>= 1
            if mask == 0:
                mask = 0x8000
            line_error -= dy
            if line_error < 0:
                y += sy
                line_error += dx

    def _draw_box_filled(self, x0, y0, x1, y1, attr):
        """Draw a filled box between the given corner points."""
        x0, y0 = self.graph_view.cutoff_coord(x0, y0)
        x1, y1 = self.graph_view.cutoff_coord(x1, y1)
        if y1 < y0:
            y0, y1 = y1, y0
        if x1 < x0:
            x0, x1 = x1, x0
        self.graph_view[y0:y1+1, x0:x1+1] = attr

    def _draw_box(self, x0, y0, x1, y1, attr, pattern=0xffff):
        """Draw an empty box between the given corner points."""
        x0, y0 = self.graph_view.cutoff_coord(x0, y0)
        x1, y1 = self.graph_view.cutoff_coord(x1, y1)
        mask = 0x8000
        mask = self._draw_straight(x1, y1, x0, y1, attr, pattern, mask)
        mask = self._draw_straight(x1, y0, x0, y0, attr, pattern, mask)
        # verticals always drawn top to bottom
        if y0 < y1:
            y0, y1 = y1, y0
        mask = self._draw_straight(x1, y1, x1, y0, attr, pattern, mask)
        mask = self._draw_straight(x0, y1, x0, y0, attr, pattern, mask)

    def _draw_straight(self, x0, y0, x1, y1, attr, pattern, mask):
        """Draw a horizontal or vertical line."""
        if x0 == x1:
            p0, p1, q, direction = y0, y1, x0, 'y'
        else:
            p0, p1, q, direction = x0, x1, y0, 'x'
        sp = 1 if p1 > p0 else -1
        for p in range(p0, p1+sp, sp):
            if pattern & mask != 0:
                if direction == 'x':
                    self.graph_view[q, p] = attr
                else:
                    self.graph_view[p, q] = attr
            mask >>= 1
            if mask == 0:
                mask = 0x8000
        return mask

    ### CIRCLE: circle, ellipse, sectors

    # NOTES ON THE MIDPOINT ALGORITHM
    #
    # CIRCLE:
    # x*x + y*y == r*r
    # look at y'=y+1
    # err(y) = y*y+x*x-r*r
    # err(y') = y*y + 2y+1 + x'*x' - r*r == err(y) + x'*x' -x*x + 2y+1
    # if x the same:
    #   err(y') == err(y) +2y+1
    # if x -> x-1:
    #   err(y') == err(y) +2y+1 -2x+1 == err(y) +2(y-x+1)
    #
    # why initialise error with 1-x == 1-r?
    # we change x if the radius is more than 0.5pix out so
    #     err(y, r+0.5) == y*y + x*x - (r*r+r+0.25) == err(y,r) - r - 0.25 >0
    # with err and r both integers, this just means
    #     err - r > 0 <==> err - r +1 >= 0
    # above, error == err(y) -r + 1 and we change x if it's >=0.
    #
    # ELLIPSE:
    # ry^2*x^2 + rx^2*y^2 == rx^2*ry^2
    # look at y'=y+1 (quadrant between points of 45deg slope)
    # err == ry^2*x^2 + rx^2*y^2 - rx^2*ry^2
    # err(y') == rx^2*(y^2+2y+1) + ry^2(x'^2)- rx^2*ry^2
    #         == err(y) + ry^2(x'^2-x^2) + rx^2*(2y+1)
    # if x the same:
    #   err(y') == err(y) + rx^2*(2y+1)
    # if x' -> x-1:
    #   err(y') == err(y) + rx^2*(2y+1) +rx^2(-2x+1)
    #
    # change x if radius more than 0.5pix out:
    #      err(y, rx+0.5, ry) == ry^2*y*y+rx^2*x*x - (ry*ry)*(rx*rx+rx+0.25) > 0
    #  ==> err(y) - (rx+0.25)*(ry*ry) > 0
    #  ==> err(y) - (rx*ry*ry + 0.25*ry*ry ) > 0
    #
    # break yinc loop if one step no longer suffices

    def circle_(self, args):
        """CIRCLE: Draw a circle, ellipse, arc or sector."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        # determine pixel aspect ratio
        pixel_aspect = (
            self._mode.pixel_height * self._screen_aspect[0],
            self._mode.pixel_width * self._screen_aspect[1]
        )
        step = next(args)
        x, y = (values.to_single(_arg).to_value() for _arg in islice(args, 2))
        r = values.to_single(next(args)).to_value()
        error.throw_if(r < 0)
        attr_index = next(args)
        if attr_index is not None:
            attr_index = values.to_int(attr_index)
        # the check is against a single precision rounded 2*pi
        check_2pi = 6.283186
        start = next(args)
        if start is not None:
            start = values.to_single(start).to_value()
            if abs(start) > check_2pi:
                raise error.BASICError(error.IFC)
        stop = next(args)
        if stop is not None:
            stop = values.to_single(stop).to_value()
            if abs(stop) > check_2pi:
                raise error.BASICError(error.IFC)
        aspect = next(args)
        if aspect is not None:
            aspect = values.to_single(aspect).to_value()
        list(args)
        x0, y0 = self._get_window_physical(x, y, step)
        if attr_index is None:
            attr_index = -1
        else:
            error.range_check(0, 255, attr_index)
        attr = self._get_attr_index(attr_index)
        if aspect is None:
            aspect = pixel_aspect[0] / float(pixel_aspect[1])
        if aspect == 1.:
            rx, _ = self._get_window_scale(r, 0.)
            ry = rx
        elif aspect > 1.:
            _, ry = self._get_window_scale(0., r)
            rx = int(round(ry / aspect))
        else:
            rx, _ = self._get_window_scale(r, 0.)
            ry = int(round(rx * aspect))
        start_octant, start_coord, start_line = -1, -1, False
        if start is not None:
            start_octant, start_coord, start_line = _get_octant(start, rx, ry)
        stop_octant, stop_coord, stop_line = -1, -1, False
        if stop is not None:
            stop_octant, stop_coord, stop_line = _get_octant(stop, rx, ry)
        if aspect == 1.:
            self._draw_circle(
                x0, y0, rx, attr,
                start_octant, start_coord, start_line,
                stop_octant, stop_coord, stop_line
            )
        else:
            startx, starty, stopx, stopy = -1, -1, -1, -1
            if start is not None:
                startx = abs(int(round(rx * math.cos(start))))
                starty = abs(int(round(ry * math.sin(start))))
            if stop is not None:
                stopx = abs(int(round(rx * math.cos(stop))))
                stopy = abs(int(round(ry * math.sin(stop))))
            self._draw_ellipse(
                x0, y0, rx, ry, attr,
                start_octant//2, startx, starty, start_line,
                stop_octant//2, stopx, stopy, stop_line
            )
        self._last_attr = attr
        self._last_point = x0, y0
        self._draw_current = None

    def _draw_circle(
            self, x0, y0, r, attr,
            oct0=-1, coo0=-1, line0=False,
            oct1=-1, coo1=-1, line1=False
        ):
        """Draw a circle sector using the midpoint algorithm."""
        # see e.g. http://en.wikipedia.org/wiki/Midpoint_circle_algorithm
        # find invisible octants
        if oct0 == -1:
            hide_oct = []
        elif oct0 < oct1 or oct0 == oct1 and _octant_gte(oct0, coo1, coo0):
            hide_oct = list(range(0, oct0)) + list(range(oct1+1, 8))
        else:
            hide_oct = list(range(oct1+1, oct0))
        # if oct1==oct0:
        # ----|.....|--- : coo1 lt coo0 : print if y in [0,coo1] or in [coo0, r]
        # ....|-----|... ; coo1 gte coo0: print if y in [coo0,coo1]
        x, y = r, 0
        bres_error = 1-r
        while x >= y:
            for octant in range(0, 8):
                if octant in hide_oct:
                    continue
                elif oct0 != oct1 and octant == oct0 and _octant_gt(oct0, coo0, y):
                    continue
                elif oct0 != oct1 and octant == oct1 and _octant_gt(oct1, y, coo1):
                    continue
                elif oct0 == oct1 and octant == oct0:
                    # if coo1 >= coo0
                    if _octant_gte(oct0, coo1, coo0):
                        # if y > coo1 or y < coo0
                        # (don't draw if y is outside coo's)
                        if _octant_gt(oct0, y, coo1) or _octant_gt(oct0, coo0,y):
                            continue
                    else:
                        # if coo0 > y > c001
                        # (don't draw if y is between coo's)
                        if _octant_gt(oct0, y, coo1) and _octant_gt(oct0, coo0, y):
                            continue
                oct_x, oct_y = _octant_coord(octant, x0, y0, x, y)
                self.graph_view[oct_y, oct_x] = attr
            # remember endpoints for pie sectors
            if y == coo0:
                coo0x = x
            if y == coo1:
                coo1x = x
            # bresenham error step
            y += 1
            if bres_error < 0:
                bres_error += 2*y+1
            else:
                x -= 1
                bres_error += 2*(y-x+1)
        # draw pie-slice lines
        if line0:
            self._draw_line(x0, y0, *_octant_coord(oct0, x0, y0, coo0x, coo0), attr=attr)
        if line1:
            self._draw_line(x0, y0, *_octant_coord(oct1, x0, y0, coo1x, coo1), attr=attr)

    def _draw_ellipse(
            self, cx, cy, rx, ry, attr,
            qua0=-1, x0=-1, y0=-1, line0=False,
            qua1=-1, x1=-1, y1=-1, line1=False
        ):
        """Draw ellipse using the midpoint algorithm."""
        # for algorithm see http://members.chello.at/~easyfilter/bresenham.html
        # find invisible quadrants
        if qua0 == -1:
            hide_qua = []
        elif qua0 < qua1 or qua0 == qua1 and _quadrant_gte(qua0, x1, y1, x0, y0):
            hide_qua = list(range(0, qua0)) + list(range(qua1+1, 4))
        else:
            hide_qua = list(range(qua1+1, qua0))
        # error increment
        dx = 16 * (1-2*rx) * ry * ry
        dy = 16 * rx * rx
        ddy = 32 * rx * rx
        ddx = 32 * ry * ry
        # error for first step
        err = dx + dy
        x, y = rx, 0
        while True:
            for quadrant in range(0,4):
                # skip invisible arc sectors
                if quadrant in hide_qua:
                    continue
                elif qua0 != qua1 and quadrant == qua0 and _quadrant_gt(qua0, x0, y0, x, y):
                    continue
                elif qua0 != qua1 and quadrant == qua1 and _quadrant_gt(qua1, x, y, x1, y1):
                    continue
                elif qua0 == qua1 and quadrant == qua0:
                    if _quadrant_gte(qua0, x1, y1, x0, y0):
                        if _quadrant_gt(qua0, x, y, x1, y1) or _quadrant_gt(qua0, x0, y0, x, y):
                            continue
                    else:
                        if _quadrant_gt(qua0, x, y, x1, y1) and _quadrant_gt(qua0, x0, y0, x, y):
                            continue
                quad_x, quad_y = _quadrant_coord(quadrant, cx, cy, x, y)
                self.graph_view[quad_y, quad_x] = attr
            # bresenham error step
            e2 = 2 * err
            if (e2 <= dy):
                y += 1
                dy += ddy
                err += dy
            if (e2 >= dx or e2 > dy):
                x -= 1
                dx += ddx
                err += dx
            # NOTE - err changes sign at the change from y increase to x increase
            if (x < 0):
                break
        # too early stop of flat vertical ellipses
        # finish tip of ellipse
        while (y < ry):
            self.graph_view[cy+y, cx] = attr
            self.graph_view[cy-y, cx] = attr
            y += 1
        # draw pie-slice lines
        if line0:
            self._draw_line(cx, cy, *_quadrant_coord(qua0, cx, cy, x0, y0), attr=attr)
        if line1:
            self._draw_line(cx, cy, *_quadrant_coord(qua1, cx, cy, x1, y1), attr=attr)

    ### PAINT: Flood fill

    def paint_(self, args):
        """PAINT: Fill an area defined by a border attribute with a tiled pattern."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        step = next(args)
        x, y = (values.to_single(_arg).to_value() for _arg in islice(args, 2))
        coord = x, y, step
        fill_attr_index, pattern = -1, None
        cval = next(args)
        if isinstance(cval, values.String):
            # pattern given; copy
            pattern = cval.to_str()
            # empty pattern "" is illegal function call
            error.throw_if(not pattern)
            # default for border, if pattern is specified as string: foreground attr
        elif cval is not None:
            fill_attr_index = values.to_int(cval)
            error.range_check(0, 255, fill_attr_index)
        border_index = next(args)
        if border_index is not None:
            border_index = values.to_int(border_index)
            error.range_check(0, 255, border_index)
        bg_pattern = next(args)
        if bg_pattern is not None:
            bg_pattern = values.pass_string(bg_pattern, err=error.IFC).to_str()
        list(args)
        # if paint *colour* specified, border default = paint colour
        # if paint *attribute* specified, border default = current foreground
        if border_index is None:
            border_index = fill_attr_index
        fill_attr = self._get_attr_index(fill_attr_index)
        border_attr = self._get_attr_index(border_index)
        self._flood_fill(coord, fill_attr, pattern, border_attr, bg_pattern)
        self._draw_current = None

    def _flood_fill(self, lcoord, fill_attr, pattern, border_attr, bg_pattern):
        """Fill an area defined by a border attribute with a tiled pattern."""
        # 4-way scanline flood fill: http://en.wikipedia.org/wiki/Flood_fill
        # flood fill stops on border colour in all directions;
        # it also stops on scanlines in fill_colour
        # pattern tiling stops at intervals that equal the pattern to be drawn,
        # unless this pattern is also equal to the background pattern.
        is_solid = (pattern is None)
        bg_tile = None
        if is_solid:
            tile = bytematrix.ByteMatrix(1, 8, fill_attr)
        else:
            tile = self._mode.build_tile(bytearray(pattern)) if pattern else None
            if bg_pattern:
                bg_tile = self._mode.build_tile(bytearray(bg_pattern))
                # only use first row of background
                bg_tile = bg_tile[:1, :]
                # illegal tile/background combo's:
                # all, or more than two consecutive, rows equal background
                for row in range(max(1, tile.height-2)):
                    comptile = tile[row:row+3, :]
                    if comptile == bg_tile.vtile(comptile.height):
                        raise error.BASICError(error.IFC)
        # viewport bounds in viewport coordinates
        bound_x0, bound_y0, bound_x1, bound_y1 = self.graph_view.get_bounds()
        x, y = self._get_window_physical(*lcoord)
        line_seed = [(x, x, y, 0)]
        # paint nothing if seed is out of bounds
        if x < bound_x0 or x > bound_x1 or y < bound_y0 or y > bound_y1:
            return
        self._last_point = x, y
        # paint nothing if we start on border attrib
        if self.graph_view[y, x] == border_attr:
            return
        while len(line_seed) > 0:
            # consider next interval
            x_start, x_stop, y, ydir = line_seed.pop()
            # extend interval as far as it goes to left and right
            x_left = x_start - self._scanline_until(border_attr, y, x_start-1, bound_x0-1).width
            x_right = x_stop + self._scanline_until(border_attr, y, x_stop+1, bound_x1+1).width
            # check next scanlines and add intervals to the list
            if ydir == 0:
                if y + 1 <= bound_y1:
                    line_seed = self._check_scanline(
                        line_seed, x_left, x_right, y+1, tile, is_solid, bg_tile, border_attr, 1
                    )
                if y - 1 >= bound_y0:
                    line_seed = self._check_scanline(
                        line_seed, x_left, x_right, y-1, tile, is_solid, bg_tile, border_attr, -1
                    )
            else:
                # check the same interval one scanline onward in the same direction
                if y+ydir <= bound_y1 and y+ydir >= bound_y0:
                    line_seed = self._check_scanline(
                        line_seed, x_left, x_right, y+ydir, tile, is_solid, bg_tile, border_attr, ydir
                    )
                # check any bit of the interval that was extended one scanline backward
                # this is where the flood fill goes around corners.
                if y-ydir <= bound_y1 and y-ydir >= bound_y0:
                    line_seed = self._check_scanline(
                        line_seed, x_left, x_start-1, y-ydir, tile, is_solid, bg_tile, border_attr, -ydir
                    )
                    line_seed = self._check_scanline(
                        line_seed, x_stop+1, x_right, y-ydir, tile, is_solid, bg_tile, border_attr, -ydir
                    )
            # draw the pixels for the current interval
            if is_solid:
                self.graph_view[y, x_left:x_right+1] = tile[0, 0]
            else:
                # convert tile to a list of attributes
                tilerow = tile[y % tile.height, :]
                n_tiles = 1 + (x_right+1) // tile.width - (x_left // tile.width)
                tiles = bytematrix.hstack((tilerow,) * n_tiles)
                interval = tiles[
                    :,
                    (x_left%tile.width) : (x_left%tile.width) + x_right - x_left + 1
                ]
                # put to screen
                self.graph_view[y, x_left:x_right+1] = interval
            # allow interrupting the paint
            if y % 4 == 0:
                self._input_methods.wait()
        self._last_attr = fill_attr

    def _scanline_until(self, element, y, x0, x1):
        """Get row until given element."""
        if x0 == x1:
            return bytematrix.ByteMatrix()
        elif x1 > x0:
            row = self.graph_view[y, x0:x1]
            try:
                # python2 won't do bytearray.index(int)
                index = row.to_bytes().index(int2byte(element))
                return row[:, :index]
            except ValueError:
                return row
        else:
            row = self.graph_view[y, x1+1:x0+1]
            try:
                index = 1 + row.to_bytes().rindex(int2byte(element))
                return row[:, index:]
            except ValueError:
                return row

    def _check_scanline(
            self, line_seed, x_start, x_stop, y,
            tile, is_solid, bg_tile, border_attr, ydir
        ):
        """Append all subintervals between border colours to the scanning stack."""
        if x_stop < x_start:
            return line_seed
        max_width = x_stop - x_start + 1
        # repeat row ceildiv + 1 times to ensure we can start in the middle of the first tile
        rtile = tile[y % tile.height, :]
        repeated_tile = rtile.htile(1 - (-max_width // rtile.width))
        if bg_tile:
            # bg_tile is only one row
            repeated_back = bg_tile.htile(1 - (-max_width // bg_tile.width))
        x = x_start
        while x <= x_stop:
            # scan horizontally until border colour found, then append interval & continue scanning
            pattern = self._scanline_until(border_attr, y, x, x_stop+1)
            if pattern.width > 0:
                # check if scanline pattern matches fill pattern
                tile_x = x % rtile.width
                has_same_pattern = (
                    # don't match zero row unless pattern is solid (special case)
                    # - avoid breaking off pattern filling on zero rows
                    # - but also don't loop forever on solid background fills
                    # - if the fill attribute is not 0, the behaviour differs:
                    #   here, the fill breaks off on encountering the matching solid line
                    (is_solid or rtile != ZERO_TILE[0, :rtile.width])
                    and pattern == repeated_tile[0, tile_x : tile_x+pattern.width]
                )
                # background tile specified: don't stop if we match the background tile (fully!)
                if bg_tile:
                    has_same_pattern = has_same_pattern and (
                        pattern.width < bg_tile.width
                        or pattern != repeated_back[0, tile_x : tile_x+pattern.width]
                    )
                # we've reached a border colour, append our interval & start a new one
                # don't append if same fill colour/pattern,
                # to avoid infinite loops over bits already painted (eg. 00 shape)
                if not has_same_pattern:
                    line_seed.append([x, x + pattern.width - 1, y, ydir])
            x += pattern.width + 1
        return line_seed

    ### PUT and GET: Sprite operations

    def put_(self, args):
        """PUT: Put a sprite on the screen."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        x0, y0 = (values.to_single(_arg).to_value() for _arg in islice(args, 2))
        array_name, operation_token = args
        array_name = self._memory.complete_name(array_name)
        operation_token = operation_token or tk.XOR
        if array_name not in self._memory.arrays:
            raise error.BASICError(error.IFC)
        elif array_name[-1:] == values.STR:
            raise error.BASICError(error.TYPE_MISMATCH)
        x0, y0 = self._get_window_physical(x0, y0)
        self._last_point = x0, y0
        packed_sprite = self._memory.arrays.view_full_buffer(array_name)
        sprite = self._mode.sprite_builder.unpack(packed_sprite)
        x1, y1 = x0 + sprite.width - 1, y0 + sprite.height - 1
        # the whole sprite must fit or it's IFC
        error.throw_if(not self.graph_view.contains(x0, y0))
        error.throw_if(not self.graph_view.contains(x1, y1))
        # apply the sprite to the screen
        if operation_token == tk.PSET:
            rect = sprite
        elif operation_token == tk.PRESET:
            rect = sprite ^ (2**self._mode.bitsperpixel - 1)
        elif operation_token == tk.AND:
            # we use in-place operations as we'll assign back anyway
            rect = operator.iand(self.graph_view[y0:y1+1, x0:x1+1], sprite)
        elif operation_token == tk.OR:
            rect = operator.ior(self.graph_view[y0:y1+1, x0:x1+1], sprite)
        elif operation_token == tk.XOR:
            rect = operator.ixor(self.graph_view[y0:y1+1, x0:x1+1], sprite)
        self.graph_view[y0:y1+1, x0:x1+1] = rect
        self._draw_current = None

    def get_(self, args):
        """GET: Read a sprite from the screen."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        x0, y0 = (values.to_single(_arg).to_value() for _arg in islice(args, 2))
        step = next(args)
        x, y = (values.to_single(_arg).to_value() for _arg in islice(args, 2))
        array_name, = args
        array_name = self._memory.complete_name(array_name)
        if array_name not in self._memory.arrays:
            raise error.BASICError(error.IFC)
        elif array_name[-1:] == values.STR:
            raise error.BASICError(error.TYPE_MISMATCH)
        x0, y0 = self._get_window_physical(x0, y0)
        self._last_point = x0, y0
        x1, y1 = self._get_window_physical(x, y, step)
        self._last_point = x1, y1
        byte_array = self._memory.arrays.view_full_buffer(array_name)
        y0, y1 = sorted((y0, y1))
        x0, x1 = sorted((x0, x1))
        # Tandy screen 6 simply GETs twice the width, it seems
        width = x1 - x0 + 1
        x1 = x0 + self._mode.sprite_builder.width_factor * width - 1
        # the whole sprite must fit or it's IFC
        error.throw_if(not self.graph_view.contains(x0, y0))
        error.throw_if(not self.graph_view.contains(x1, y1))
        # set size record
        # read from screen and convert to byte array
        sprite = self.graph_view[y0:y1+1, x0:x1+1]
        packed_sprite = self._mode.sprite_builder.pack(sprite)
        try:
            byte_array[:len(packed_sprite)] = packed_sprite
        except ValueError:
            # cannot modify size of memoryview object - sprite larger than array
            raise error.BASICError(error.IFC)
        self._draw_current = None

    ### DRAW statement

    def draw_(self, args):
        """DRAW: Execute a Graphics Macro Language string."""
        if self._mode.is_text_mode:
            raise error.BASICError(error.IFC)
        gml = values.next_string(args)
        self._draw(gml)
        list(args)

    def _draw(self, gml):
        """Execute a Graphics Macro Language string."""
        # don't convert to uppercase as VARPTR$ elements are case sensitive
        gmls = mlparser.MLParser(gml, self._memory, self._values)
        if not self._draw_current:
            self._draw_current = self._last_point
        plot, goback = True, False
        while True:
            c = gmls.skip_blank_read().upper()
            if c == b'':
                break
            elif c == b';':
                continue
            elif c == b'B':
                # do not draw
                plot = False
            elif c == b'N':
                # return to postiton after move
                goback = True
            elif c == b'X':
                # execute substring
                sub = gmls.parse_string()
                self._draw(sub)
            elif c == b'C':
                # set foreground colour
                # allow empty spec (default 0), but only if followed by a semicolon
                if gmls.skip_blank() == b';':
                    self._last_attr = 0
                else:
                    attr = gmls.parse_number()
                    # 100000 seems to be GW's limit
                    error.range_check(-99999, 99999, attr)
                    self._last_attr = attr
            elif c == b'S':
                # set scale
                scale = gmls.parse_number()
                error.range_check(1, 255, scale)
                self._draw_scale = scale
            elif c == b'A':
                # set angle
                # allow empty spec (default 0), but only if followed by a semicolon
                if gmls.skip_blank() == b';':
                    self._draw_angle = 0
                else:
                    angle = gmls.parse_number()
                    error.range_check(0, 3, angle)
                    self._draw_angle = 90 * angle
            elif c == b'T':
                # 'turn angle' - set (don't turn) the angle to any value
                if gmls.read(1).upper() != b'A':
                    raise error.BASICError(error.IFC)
                # allow empty spec (default 0), but only if followed by a semicolon
                if gmls.skip_blank() == b';':
                    self._draw_angle = 0
                else:
                    angle = gmls.parse_number()
                    error.range_check(-360, 360, angle)
                    self._draw_angle = angle
            # one-variable movement commands:
            elif c in (b'U', b'D', b'L', b'R', b'E', b'F', b'G', b'H'):
                step = gmls.parse_number(default=1)
                # 100000 seems to be GW's limit
                error.range_check(-99999, 99999, step)
                x0, y0 = self._draw_current
                x1, y1 = 0, 0
                if c in (b'U', b'E', b'H'):
                    y1 -= step
                elif c in (b'D', b'F', b'G'):
                    y1 += step
                if c in (b'L', b'G', b'H'):
                    x1 -= step
                elif c in (b'R', b'E', b'F'):
                    x1 += step
                self._draw_step(x0, y0, x1, y1, plot, goback)
                plot = True
                goback = False
            # two-variable movement command
            elif c == b'M':
                relative = gmls.skip_blank() in (b'+', b'-')
                x = gmls.parse_number()
                error.range_check(-9999, 9999, x)
                if gmls.skip_blank() != b',':
                    raise error.BASICError(error.IFC)
                else:
                    gmls.read(1)
                y = gmls.parse_number()
                error.range_check(-9999, 9999, y)
                x0, y0 = self._draw_current
                if relative:
                    self._draw_step(x0, y0, x, y, plot, goback)
                else:
                    if plot:
                        self._draw_line(x0, y0, x, y, self._last_attr)
                    self._draw_current = x, y
                    if goback:
                        self._draw_current = x0, y0
                plot = True
                goback = False
            elif c == b'P':
                # paint - flood fill
                fill_idx = gmls.parse_number()
                error.range_check(0, 9999, fill_idx)
                if gmls.skip_blank_read() != b',':
                    raise error.BASICError(error.IFC)
                border_idx = gmls.parse_number()
                error.range_check(0, 9999, border_idx)
                x, y = self._get_window_logical(*self._draw_current)
                fill_attr = self._get_attr_index(fill_idx)
                border_attr = self._get_attr_index(border_idx)
                self._flood_fill((x, y, False), fill_attr, None, border_attr, None)
            else:
                raise error.BASICError(error.IFC)
        # if WINDOW is set, the current position for non-DRAW commands does not track
        if self._window_bounds is None:
            self._last_point = self._draw_current

    def _draw_step(self, x0, y0, sx, sy, plot, goback):
        """Make a DRAW step, drawing a line and returning if requested."""
        scale = self._draw_scale
        rotate = self._draw_angle
        # pixel aspect ratio
        aspect = (
            self._mode.pixel_height * self._screen_aspect[0],
            self._mode.pixel_width * self._screen_aspect[1]
        )
        yfac = float(aspect[1]) / float(aspect[0])
        x1 = int(math.trunc(scale*sx / 4.))
        y1 = int(math.trunc(scale*sy / 4.))
        if rotate == 0 or rotate == 360:
            pass
        elif rotate == 90:
            x1, y1 = int(y1*yfac), -int(x1//yfac)
        elif rotate == 180:
            x1, y1 = -x1, -y1
        elif rotate == 270:
            x1, y1 = -int(y1*yfac), int(x1//yfac)
        else:
            fx, fy = float(x1), float(y1)
            # degrees to radians
            phi = rotate * math.pi / 180.
            sinr, cosr = math.sin(phi), math.cos(phi)
            x1 = cosr * fx + sinr*fy * yfac
            y1 = cosr * fy - sinr*fx / yfac
            x1, y1 = int(round(x1)), int(round(y1))
        y1 += y0
        x1 += x0
        if plot:
            self._draw_line(x0, y0, x1, y1, self._last_attr)
        if goback:
            self._draw_current = x0, y0
        else:
            self._draw_current = x1, y1

    ### POINT and PMAP

    def point_(self, args):
        """
        POINT (1 argument): Return current coordinate
        (2 arguments): Return the attribute of a pixel.
        """
        arg0 = next(args)
        arg1 = next(args)
        if arg1 is None:
            arg0 = values.to_integer(arg0)
            fn = values.to_int(arg0)
            error.range_check(0, 3, fn)
            list(args)
            if self._mode.is_text_mode:
                return self._values.new_single()
            # if DRAW and other commands are out of sync, POINT is adjusted to the latest
            # (if non-DRAW commands were executed, draw_current is set to None)
            current = self._draw_current or self._last_point
            if fn in (0, 1):
                point = current[fn]
            elif fn in (2, 3):
                point = self._get_window_logical(*current)[fn - 2]
            return self._values.new_single().from_value(point)
        else:
            if self._mode.is_text_mode:
                raise error.BASICError(error.IFC)
            arg1 = values.pass_number(arg1)
            list(args)
            x, y = values.to_single(arg0).to_value(), values.to_single(arg1).to_value()
            x, y = self._get_window_physical(x, y)
            if x < 0 or x >= self._mode.pixel_width or y < 0 or y >= self._mode.pixel_height:
                point = -1
            else:
                point = self.graph_view[y, x]
            return self._values.new_integer().from_int(point)

    def pmap_(self, args):
        """PMAP: convert between logical and physical coordinates."""
        # create a new Single for the return value
        coord = values.to_single(next(args))
        mode = values.to_integer(next(args))
        list(args)
        mode = mode.to_int()
        error.range_check(0, 3, mode)
        if self._mode.is_text_mode:
            if mode in (2, 3):
                values.to_integer(coord)
            value = 0
        elif mode == 0:
            value, _ = self._get_window_physical(values.to_single(coord).to_value(), 0.)
        elif mode == 1:
            _, value = self._get_window_physical(0., values.to_single(coord).to_value())
        elif mode == 2:
            value, _ = self._get_window_logical(values.to_integer(coord).to_int(), 0)
        elif mode == 3:
            _, value = self._get_window_logical(0, values.to_integer(coord).to_int())
        return self._values.new_single().from_value(value)



###############################################################################
# octant logic for CIRCLE

def _get_octant(f, rx, ry):
    """Get the circle octant for a given coordinate."""
    neg = f < 0.
    f = abs(f)
    octant = int(f / (math.pi / 4.))
    # deal with values that should equal 2*pi but are a bit bigger due to Single precision math
    if octant == 8:
        octant = 7
    if octant in (0, 3, 4, 7):
        # running var is y
        coord = abs(int(round(ry * math.sin(f))))
    else:
        # running var is x
        coord = abs(int(round(rx * math.cos(f))))
    return octant, coord, neg

def _octant_coord(octant, x0, y0, x, y):
    """Return symmetrically reflected coordinates for a given pair."""
    if octant == 7:
        return x0+x, y0+y
    elif octant == 0:
        return x0+x, y0-y
    elif octant == 4:
        return x0-x, y0+y
    elif octant == 3:
        return x0-x, y0-y
    elif octant == 6:
        return x0+y, y0+x
    elif octant == 1:
        return x0+y, y0-x
    elif octant == 5:
        return x0-y, y0+x
    elif octant == 2:
        return x0-y, y0-x

def _octant_gt(octant, y, coord):
    """Return whether y is further along the circle than coord."""
    if octant%2 == 1:
        return y < coord
    else:
        return y > coord

def _octant_gte(octant, y, coord):
    """Return whether y is further along the circle than coord, or equal."""
    if octant%2 == 1:
        return y <= coord
    else:
        return y >= coord


###############################################################################
# quadrant logic for CIRCLE

def _quadrant_coord(quadrant, x0, y0, x, y):
    """Return symmetrically reflected coordinates for a given pair."""
    if quadrant == 3:
        return x0+x, y0+y
    elif quadrant == 0:
        return x0+x, y0-y
    elif quadrant == 2:
        return x0-x, y0+y
    elif quadrant == 1:
        return x0-x, y0-y

def _quadrant_gt(quadrant, x, y, x0, y0):
    """Return whether y is further along the ellipse than coord."""
    if quadrant%2 == 0:
        if y != y0:
            return y > y0
        else:
            return x < x0
    else:
        if y != y0:
            return y < y0
        else:
            return x > x0

def _quadrant_gte(quadrant, x, y, x0, y0):
    """Return whether y is further along the ellipse than coord, or equal."""
    if quadrant%2 == 0:
        if y != y0:
            return y > y0
        else:
            return x <= x0
    else:
        if y != y0:
            return y < y0
        else:
            return x >= x0



"""
PC-BASIC - basicevents.py
Handlers for BASIC events

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from contextlib import contextmanager

from .base import scancode
from .base import error
from .base import tokens as tk
from .base import signals
from . import values


###############################################################################
# BASIC events

class BasicEvents(object):
    """Manage BASIC events."""

    def __init__(self, sound, clock, files, program, num_fn_keys, tandy_fn_keys):
        """Initialise event triggers."""
        self._sound = sound
        self._clock = clock
        # files for com1 and com2
        self._files = files
        # for on_event_gosub_
        self._program = program
        # 10 or 12 definable function keys
        self._num_fn_keys = num_fn_keys
        # key codes are shifted by 2 on Tandy
        self._tandy_fn_keys = tandy_fn_keys
        self.reset()

    def reset(self):
        """Reset event triggers."""
        # KEY: init key events
        keys = [
            scancode.F1, scancode.F2, scancode.F3, scancode.F4, scancode.F5,
            scancode.F6, scancode.F7, scancode.F8, scancode.F9, scancode.F10]
        # on late IBM BASICA and BASICJ versions:
        # * KEY 30 and KEY 31 refer to F11 and F12, as in QBASIC
        # * there are 10 definable functions 15-25
        # on classic GW-BASIC:
        # * F11 and F12 are not accessible
        # * the definable range is 15-20
        # on Tandy:
        # * F11 and F12 are accessible trough keys 11 and 12,
        # * the arrow keys codes are shifted by 2
        # * the definable range is 17-20
        if self._tandy_fn_keys:
            # Tandy only
            keys += [scancode.F11, scancode.F12]
            keys += [scancode.UP, scancode.LEFT, scancode.RIGHT, scancode.DOWN]
            keys += [None] * 4
        elif self._num_fn_keys == 12:
            # keys 11-14
            keys += [scancode.UP, scancode.LEFT, scancode.RIGHT, scancode.DOWN]
            # keys 15-25
            keys += [None] * 11
            # keys 26-29 - these should not be accessible but are
            keys += [None] * 4
            # non-Tandy F11 and F12 mapped to 30, 31
            keys += [scancode.F11, scancode.F12]
        else:
            keys += [scancode.UP, scancode.LEFT, scancode.RIGHT, scancode.DOWN]
            keys += [None] * 6
        self.key = [KeyHandler(sc) for sc in keys]
        # other events
        self.timer = TimerHandler(self._clock)
        self.play = PlayHandler(self._sound)
        self.com = [
            ComHandler(self._files.get_device(b'COM1:')),
            ComHandler(self._files.get_device(b'COM2:'))]
        self.pen = PenHandler()
        # joy*2 + button
        self.strig = [
            StrigHandler(joy, button)
            for joy in range(2) for button in range(2)
        ]
        # key events are not handled FIFO but first arrow keys and definable keys, then function keys
        if self._tandy_fn_keys:
            ordered_keys = self.key[12:] + self.key[:12]
        elif self._num_fn_keys == 12:
            ordered_keys = self.key[10:29] + self.key[:10] + self.key[29:]
        else:
            ordered_keys = self.key[10:] + self.key[:10]
        # all handlers in order of handling; TIMER first
        self.all = [self.timer] + ordered_keys + [self.play] + self.com + [self.pen] + self.strig
        # keep a list of enabled events
        self.enabled = set()
        # set suspension off
        self.suspend_all = False

    def command(self, handler, command_char):
        """Turn the event ON, OFF and STOP."""
        if command_char == tk.ON:
            self.enabled.add(handler)
            handler.stopped = False
        elif command_char == tk.OFF:
            # we seem to need to keep ComHandler around to make serial events work correctly
            # i.e. apparently they can be triggered when switched off - I don't understand why
            if not isinstance(handler, ComHandler):
                self.enabled.discard(handler)
        elif command_char == tk.STOP:
            handler.stopped = True


    ##########################################################################
    # callbacks

    def pen_(self, args):
        """PEN: switch on/off light pen event handling."""
        command, = args
        self.command(self.pen, command)

    def strig_(self, args):
        """STRIG: switch on/off fire button event handling."""
        num = values.to_int(next(args))
        command, = args
        error.range_check(0, 255, num)
        if num in (0, 2, 4, 6):
            self.command(self.strig[num//2], command)

    def com_(self, args):
        """COM: switch on/off serial port event handling."""
        num = values.to_int(next(args))
        command, = args
        error.range_check(0, 2, num)
        if num > 0:
            self.command(self.com[num-1], command)

    def timer_(self, args):
        """TIMER: switch on/off timer event handling."""
        command, = args
        self.command(self.timer, command)

    def key_(self, args):
        """KEY: switch on/off keyboard events."""
        num = values.to_int(next(args))
        error.range_check(0, 255, num)
        command, = args
        # others are ignored
        if num >= 1 and num <= 20:
            self.command(self.key[num-1], command)

    def play_(self, args):
        """PLAY: switch on/off sound queue event handling."""
        command, = args
        self.command(self.play, command)

    def on_event_gosub_(self, args):
        """ON .. GOSUB: define event trapping subroutine."""
        token = next(args)
        num = next(args)
        jumpnum = next(args)
        if jumpnum == 0:
            jumpnum = None
        elif jumpnum not in self._program.line_numbers:
            raise error.BASICError(error.UNDEFINED_LINE_NUMBER)
        list(args)
        if token == tk.KEY:
            keynum = values.to_int(num)
            error.range_check(1, len(self.key), keynum)
            self.key[keynum-1].set_jump(jumpnum)
        elif token == tk.TIMER:
            timeval = values.to_single(num).to_value()
            error.throw_if(timeval <= 0)
            period = round(timeval * 1000.)
            self.timer.set_trigger(period)
            self.timer.set_jump(jumpnum)
        elif token == tk.PLAY:
            playval = values.to_int(num)
            error.range_check(1, 32, playval)
            self.play.set_trigger(playval)
            self.play.set_jump(jumpnum)
        elif token == tk.PEN:
            self.pen.set_jump(jumpnum)
        elif token == tk.STRIG:
            strigval = values.to_int(num)
            ## 0 -> [0][0] 2 -> [0][1]  4-> [1][0]  6 -> [1][1]
            if strigval not in (0,2,4,6):
                raise error.BASICError(error.IFC)
            self.strig[strigval//2].set_jump(jumpnum)
        elif token == tk.COM:
            comnum = values.to_int(num)
            error.range_check(1, 2, comnum)
            self.com[comnum-1].set_jump(jumpnum)


class EventHandler(object):
    """Manage event triggers."""

    def __init__(self):
        """Initialise untriggered and disabled."""
        self.reset()

    def reset(self):
        """Reset to untriggered and disabled initial state."""
        self.gosub = None
        self.stopped = False
        self.triggered = False

    def set_jump(self, jump):
        """Set the jump line number."""
        self.gosub = jump

    def trigger(self):
        """Trigger the event."""
        self.triggered = True

    def check_input(self, signal):
        """Stub for event checker."""
        return False


class PlayHandler(EventHandler):
    """Manage PLAY (music queue) events."""

    def __init__(self, sound):
        """Initialise PLAY trigger."""
        EventHandler.__init__(self)
        self.trig = 1
        self._sound = sound
        # set to a number higher than the maximum buffer length?
        self.last = 0 #34 if multivoice else 0

    def check_input(self, signal):
        """Check and trigger PLAY (music queue) events."""
        play_now = self._sound.tones_waiting()
        if self._sound.multivoice:
            if (self.last > play_now and play_now < self.trig):
                self.trigger()
        else:
            if (self.last >= self.trig and play_now < self.trig):
                self.trigger()
        self.last = play_now
        return False

    def set_trigger(self, n):
        """Set PLAY trigger to n notes."""
        self.trig = n


class TimerHandler(EventHandler):
    """Manage TIMER events."""

    def __init__(self, clock):
        """Initialise TIMER trigger."""
        EventHandler.__init__(self)
        self.period = 0
        self.start = 0
        self.clock = clock

    def set_trigger(self, n):
        """Set TIMER trigger to n milliseconds."""
        self.start = self.clock.get_time_ms()
        self.period = n

    def check_input(self, signal):
        """Trigger TIMER events."""
        mutimer = self.clock.get_time_ms()
        if mutimer >= self.start + self.period:
            self.start = mutimer
            self.trigger()
        return False


class ComHandler(EventHandler):
    """Manage COM-port events."""

    def __init__(self, com_device):
        """Initialise COM trigger."""
        EventHandler.__init__(self)
        self.device = com_device

    # treat com-port "trigger" as real-time check

    @property
    def triggered(self):
        return self.device.char_waiting()

    @triggered.setter
    def triggered(self, value):
        pass


class KeyHandler(EventHandler):
    """Manage KEY events."""

    # build KEY trigger code
    # see http://www.petesqbsite.com/sections/tutorials/tuts/keysdet.txt
    # second byte is scan code; first byte
    #  0       if the key is pressed alone
    #  1 to 3    if any Shift and the key are combined
    #    4       if Ctrl and the key are combined
    #    8       if Alt and the key are combined
    #   32       if NumLock is activated
    #   64       if CapsLock is activated
    #  128       if we are defining some extended key
    #
    # "extended keys" are the additional keys brought in with the 101-key Type M keyboard
    # for example the arrow keys, ins, del, home, end, pgp, pgdn on the *non-numerical* keyboard
    #
    # for predefined keys, modifier is ignored

    # modifier flags, merging left-shift and right-shift and excluding the "extended" flag
    # we cannot reliably detect whether a key supplied by the interface is "extended".
    _mod_flags = {
        scancode.CAPSLOCK: 0x40, scancode.NUMLOCK: 0x20,
        scancode.ALT: 0x8, scancode.CTRL: 0x4,
        # both shifts are equal for event purposes
        scancode.LSHIFT: 0x3, scancode.RSHIFT: 0x3,
    }

    def __init__(self, scancode=None):
        """Initialise KEY trigger."""
        EventHandler.__init__(self)
        self._modcode = None
        self._scancode = scancode
        self._predefined = (scancode is not None)

    def check_input(self, signal):
        """Trigger KEY events."""
        if (self._scancode is not None) and (signal.event_type == signals.KEYB_DOWN):
            _, scancode, modifiers = signal.params
            if scancode != self._scancode:
                return False
            modcode = 0
            if modifiers and not self._predefined:
                for m in modifiers:
                    modcode |= self._mod_flags.get(m, 0)
            if self._predefined or self._modcode == modcode:
                # trigger event
                self.trigger()
                # drop key from key buffer
                # True removes signal from further processing
                return True
        return False

    def set_trigger(self, keystr):
        """Set KEY trigger to chr(modcode)+chr(scancode)."""
        # only length-2 expressions can be assigned as scancode triggers
        if len(keystr) != 2:
            raise error.BASICError(error.IFC)
        # can't redefine scancodes for predefined keys 1-14 (pc) 1-16 (tandy)
        if not self._predefined:
            # some modifier codes are different from the ones used internally & seen in peek(1047):
            #
            # value | on key   | peek (1047)
            # ------|----------|------------
            # 0x80  | Extended | Insert
            # 0x10  | not used | Scroll Lock
            #
            # exclude 0x10 and 0x80 from the modifier mask
            # this means we're ignoring the "extended" flag
            self._modcode = bytearray(keystr)[0] & 0x6f
            self._scancode = bytearray(keystr)[1]
            # per the docs, &h02 is left-shift and &h01 is right-shift
            # in reality, all shifts are equal
            if self._modcode & 3:
                self._modcode |= 3


class PenHandler(EventHandler):
    """Manage PEN events."""

    def __init__(self):
        """Initialise STRIG trigger."""
        EventHandler.__init__(self)

    def check_input(self, signal):
        """Trigger PEN events."""
        if signal.event_type == signals.PEN_DOWN:
            self.trigger()
        # don't swallow event
        return False


class StrigHandler(EventHandler):
    """Manage STRIG events."""

    def __init__(self, joy, button):
        """Initialise STRIG trigger."""
        EventHandler.__init__(self)
        self._joybutton = joy, button

    def check_input(self, signal):
        """Trigger STRIG events."""
        if (signal.event_type == signals.STICK_DOWN) and (signal.params == self._joybutton):
            self.trigger()



"""
PC-BASIC - application data package
Fonts, codepages, bundled programs and application branding

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import json as _json

from ..compat import resources as _resources

from .programs import PROGRAMS, read_program_file
from .fonts import FONTS, read_fonts
from .codepages import CODEPAGES, read_codepage

ICON = tuple(_json.loads(_resources.read_binary(__package__, 'icon.json')))



"""
PC-BASIC - expressions.py
Expression stack

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from collections import deque
from functools import partial
import logging
import struct
import types

from ..base import tokens as tk
from ..base.tokens import DIGITS, LETTERS
from ..base import error
from .. import values
from .. import dos
from . import operators as op
from . import userfunctions


class ExpressionParser(object):
    """Expression parser."""

    def __init__(self, values, memory):
        """Initialise empty expression."""
        self._values = values
        # for variable retrieval
        self._memory = memory
        # user-defined functions
        self.user_functions = userfunctions.UserFunctionManager(memory, values, self)
        # initialise syntax tables
        self._init_syntax()
        # callbacks must be initilised later
        self._callbacks = {}
        self._extensions = {}

    def _init_syntax(self):
        """Initialise function syntax tables."""
        self._complex = {
            tk.USR: {
                None: self._gen_parse_arguments,
                tk.C_0: self._gen_parse_arguments,
                tk.C_1: self._gen_parse_arguments,
                tk.C_2: self._gen_parse_arguments,
                tk.C_3: self._gen_parse_arguments,
                tk.C_4: self._gen_parse_arguments,
                tk.C_5: self._gen_parse_arguments,
                tk.C_6: self._gen_parse_arguments,
                tk.C_7: self._gen_parse_arguments,
                tk.C_8: self._gen_parse_arguments,
                tk.C_9: self._gen_parse_arguments,
            },
            tk.IOCTL: {
                b'$': self._gen_parse_ioctl,
            },
            tk.ENVIRON: {
                b'$': self._gen_parse_arguments,
            },
            tk.INPUT: {
                b'$': self._gen_parse_input,
            },
            tk.ERDEV: {
                b'$': self._no_argument,
                None: self._no_argument,
            },
            tk.VARPTR: {
                b'$': self._gen_parse_varptr_str,
                None: self._gen_parse_varptr,
            },
        }
        self._simple = {
            tk.ERL: self._no_argument,
            tk.ERR: self._no_argument,
            tk.CSRLIN: self._no_argument,
            tk.INKEY: self._no_argument,
            tk.DATE: self._no_argument,
            tk.TIME: self._no_argument,
            tk.TIMER: self._no_argument,
            tk.RND: self._gen_parse_one_optional_argument,
            tk.CVI: self._gen_parse_arguments,
            tk.CVS: self._gen_parse_arguments,
            tk.CVD: self._gen_parse_arguments,
            tk.MKI: self._gen_parse_arguments,
            tk.MKS: self._gen_parse_arguments,
            tk.MKD: self._gen_parse_arguments,
            tk.SGN: self._gen_parse_arguments,
            tk.INT: self._gen_parse_arguments,
            tk.FIX: self._gen_parse_arguments,
            tk.ABS: self._gen_parse_arguments,
            tk.SQR: self._gen_parse_arguments,
            tk.SIN: self._gen_parse_arguments,
            tk.LOG: self._gen_parse_arguments,
            tk.EXP: self._gen_parse_arguments,
            tk.COS: self._gen_parse_arguments,
            tk.TAN: self._gen_parse_arguments,
            tk.ATN: self._gen_parse_arguments,
            tk.PEEK: self._gen_parse_arguments,
            tk.FRE: self._gen_parse_arguments,
            tk.INP: self._gen_parse_arguments,
            tk.POS: self._gen_parse_arguments,
            tk.CINT: self._gen_parse_arguments,
            tk.CSNG: self._gen_parse_arguments,
            tk.CDBL: self._gen_parse_arguments,
            tk.LEN: self._gen_parse_arguments,
            tk.STR: self._gen_parse_arguments,
            tk.VAL: self._gen_parse_arguments,
            tk.ASC: self._gen_parse_arguments,
            tk.CHR: self._gen_parse_arguments,
            tk.SPACE: self._gen_parse_arguments,
            tk.OCT: self._gen_parse_arguments,
            tk.HEX: self._gen_parse_arguments,
            tk.PEN: self._gen_parse_arguments,
            tk.STICK: self._gen_parse_arguments,
            tk.STRIG: self._gen_parse_arguments,
            tk.EOF: self._gen_parse_arguments,
            tk.LOC: self._gen_parse_arguments,
            tk.LOF: self._gen_parse_arguments,
            tk.LPOS: self._gen_parse_arguments,
            tk.EXTERR: self._gen_parse_arguments,
            tk.PLAY: self._gen_parse_arguments,
            tk.STRING: partial(self._gen_parse_arguments, length=2),
            tk.PMAP: partial(self._gen_parse_arguments, length=2),
            tk.LEFT: partial(self._gen_parse_arguments, length=2),
            tk.RIGHT: partial(self._gen_parse_arguments, length=2),
            tk.POINT: partial(self._gen_parse_arguments_optional, length=2),
            tk.MID: partial(self._gen_parse_arguments_optional, length=3),
            tk.SCREEN: partial(self._gen_parse_arguments_optional, length=3),
            tk.INSTR: self._gen_parse_instr,
            tk.FN: None,
            b'_': self._gen_parse_call_extension,
        }
        self._functions = set(self._complex.keys()) | set(self._simple.keys())

    def init_functions(self, session):
        """Initialise function callbacks."""
        self._callbacks = {
            tk.USR: session.machine.usr_,
            tk.USR + tk.C_0: session.machine.usr_,
            tk.USR + tk.C_1: session.machine.usr_,
            tk.USR + tk.C_2: session.machine.usr_,
            tk.USR + tk.C_3: session.machine.usr_,
            tk.USR + tk.C_4: session.machine.usr_,
            tk.USR + tk.C_5: session.machine.usr_,
            tk.USR + tk.C_6: session.machine.usr_,
            tk.USR + tk.C_7: session.machine.usr_,
            tk.USR + tk.C_8: session.machine.usr_,
            tk.USR + tk.C_9: session.machine.usr_,
            tk.IOCTL + b'$': session.files.ioctl_,
            tk.ENVIRON + b'$': session.environment.environ_,
            tk.INPUT + b'$': session.files.input_,
            tk.ERDEV: session.files.erdev_,
            tk.ERDEV + b'$': session.files.erdev_str_,
            tk.VARPTR: session.memory.varptr_,
            tk.VARPTR + b'$': session.memory.varptr_str_,
            tk.SCREEN: session.text_screen.screen_fn_,
            tk.FN: None,
            tk.ERL: session.interpreter.erl_,
            tk.ERR: session.interpreter.err_,
            tk.STRING: values.string_,
            tk.INSTR: values.instr_,
            tk.CSRLIN: session.text_screen.csrlin_,
            tk.POINT: session.graphics.point_,
            tk.INKEY: session.keyboard.inkey_,
            tk.CVI: values.cvi_,
            tk.CVS: values.cvs_,
            tk.CVD: values.cvd_,
            tk.MKI: values.mki_,
            tk.MKS: values.mks_,
            tk.MKD: values.mkd_,
            tk.EXTERR: session.files.exterr_,
            tk.DATE: session.clock.date_fn_,
            tk.TIME: session.clock.time_fn_,
            tk.PLAY: session.sound.play_fn_,
            tk.TIMER: session.clock.timer_,
            tk.PMAP: session.graphics.pmap_,
            tk.LEFT: values.left_,
            tk.RIGHT: values.right_,
            tk.MID: values.mid_,
            tk.SGN: values.sgn_,
            tk.INT: values.int_,
            tk.ABS: values.abs_,
            tk.SQR: values.sqr_,
            tk.RND: session.randomiser.rnd_,
            tk.SIN: values.sin_,
            tk.LOG: values.log_,
            tk.EXP: values.exp_,
            tk.COS: values.cos_,
            tk.TAN: values.tan_,
            tk.ATN: values.atn_,
            tk.FRE: session.memory.fre_,
            tk.INP: session.machine.inp_,
            tk.POS: session.text_screen.pos_,
            tk.LEN: values.len_,
            tk.STR: values.str_,
            tk.VAL: values.val_,
            tk.ASC: values.asc_,
            tk.CHR: values.chr_,
            tk.PEEK: session.all_memory.peek_,
            tk.SPACE: values.space_,
            tk.OCT: values.oct_,
            tk.HEX: values.hex_,
            tk.LPOS: session.files.lpos_,
            tk.CINT: values.cint_,
            tk.CSNG: values.csng_,
            tk.CDBL: values.cdbl_,
            tk.FIX: values.fix_,
            tk.PEN: session.pen_fn_,
            tk.STICK: session.stick.stick_,
            tk.STRIG: session.stick.strig_,
            tk.EOF: session.files.eof_,
            tk.LOC: session.files.loc_,
            tk.LOF: session.files.lof_,
            b'_': session.extensions.call_as_function,
        }

    def __getstate__(self):
        """Pickle."""
        pickle_dict = self.__dict__.copy()
        # functools.partial objects and functions can't be pickled
        pickle_dict['_simple'] = None
        pickle_dict['_complex'] = None
        pickle_dict['_callbacks'] = None
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Unpickle."""
        self.__dict__.update(pickle_dict)
        self._init_syntax()

    def parse_expression(self, ins):
        """Parse and evaluate tokenised expression."""
        self._memory.strings.reset_temporaries()
        return self.parse(ins)

    def parse(self, ins):
        """Parse and evaluate tokenised (sub-)expression."""
        operations = deque()
        with self._memory.get_stack() as units:
            final = True
            # see https://en.wikipedia.org/wiki/Shunting-yard_algorithm
            d = b''
            while True:
                last = d
                ins.skip_blank()
                d = ins.read_keyword_token()
                ins.seek(-len(d), 1)
                if d == tk.NOT and not (last in op.OPERATORS or last == b''):
                    # unary NOT ends expression except after another operator or at start
                    break
                elif d in op.OPERATORS:
                    ins.read(len(d))
                    # get combined operators such as >=
                    if d in op.COMBINABLE:
                        nxt = ins.skip_blank()
                        if nxt in op.COMBINABLE:
                            d += ins.read(len(nxt))
                    if last in op.OPERATORS or last == b'' or d == tk.NOT:
                        # also if last is ( but that leads to recursive call and last == ''
                        nargs = 1
                        # zero operands for a binary operator is always syntax error
                        # because it will be seen as an illegal unary
                        try:
                            oper = op.UNARY[d]
                            prec = op.PRECEDENCE[(d, nargs)]
                        except KeyError:
                            raise error.BASICError(error.STX)
                    else:
                        nargs = 2
                        try:
                            oper = op.BINARY[d]
                            prec = op.PRECEDENCE[(d, nargs)]
                        except KeyError:
                            # illegal combined ops like == raise syntax error here
                            raise error.BASICError(error.STX)
                        self._drain(prec, operations, units)
                    operations.append((oper, nargs, prec))
                elif not (last in op.OPERATORS or last == b''):
                    # repeated unit ends expression
                    # repeated literals or variables or non-keywords like 'AS'
                    break
                elif d == b'(':
                    ins.read(len(d))
                    # we need to create a new object or we'll overwrite our own stacks
                    # this will not be needed if we localise stacks in the expression parser
                    # either a separate class of just as local variables
                    units.append(self.parse(ins))
                    ins.require_read((b')',))
                elif d and d in LETTERS:
                    name = ins.read_name()
                    error.throw_if(not name, error.STX)
                    indices = self.parse_indices(ins)
                    view = self._memory.view_or_create_variable(name, indices)
                    # should make a shallow copy? but .clone here breaks circular MID$
                    units.append(view)
                elif d in self._functions:
                    units.append(self._parse_function(ins, d))
                    #if not isinstance(units[-1], values.String):
                    #    self._memory.strings.reset_temporaries()
                elif d in tk.END_STATEMENT:
                    break
                elif d in tk.END_EXPRESSION:
                    # missing operand inside brackets or before comma is syntax error
                    final = False
                    break
                elif d == b'"':
                    units.append(self.read_string_literal(ins))
                else:
                    units.append(self.read_number_literal(ins))
            # raises IndexError for insufficient operators
            try:
                self._drain(0, operations, units)
                return units[0]
            except IndexError:
                # empty expression is a syntax error (inside brackets)
                # or Missing Operand (in an assignment)
                if final:
                    raise error.BASICError(error.MISSING_OPERAND)
                raise error.BASICError(error.STX)

    def _drain(self, precedence, operations, units):
        """Drain evaluation stack until an operator of low precedence on top."""
        while operations:
            # this raises IndexError if there are not enough operators
            if precedence > operations[-1][2]:
                break
            oper, narity, _ = operations.pop()
            args = reversed([units.pop() for _ in range(narity)])
            units.append(oper(*args))

    def read_string_literal(self, ins):
        """Read a quoted string literal (no leading blanks), return as String."""
        # address points to initial quote
        address = ins.tell_address()
        value = ins.read_string().strip(b'"')
        # if this is a program, create a string pointer to code space
        # and don't reserve space in string memory
        # +1 to point to start of payload, not intial quote
        return self._values.from_str_at(value, None if address is None else address + 1)

    def read_number_literal(self, ins):
        """Return the value of a numeric literal (no leading blanks)."""
        d = ins.peek()
        # number literals as ASCII are accepted in tokenised streams. only if they start with a figure (not & or .)
        # this happens e.g. after non-keywords like AS. They are not acceptable as line numbers.
        if d in DIGITS:
            return self._values.from_repr(ins.read_number(), allow_nonnum=False)
        # number literals
        elif d in tk.NUMBER:
            return self._values.from_token(ins.read_number_token())
        elif d == tk.T_UINT:
            # gw-basic allows adding line numbers to numbers
            # drop 0E token, interpret payload to unsigned integer
            value = struct.unpack('<bH', ins.read(3))[1]
            # we need to convert to single to ensure it is interpreted as the unsigned value
            return self._values.new_single().from_int(value)
        else:
            raise error.BASICError(error.STX)

    def parse_indices(self, ins):
        """Parse array indices."""
        indices = []
        if ins.skip_blank_read_if((b'[', b'(')):
            # it's an array, read indices
            while True:
                expr = self.parse(ins)
                indices.append(values.to_int(expr))
                if not ins.skip_blank_read_if((b',',)):
                    break
            ins.require_read((b']', b')'))
        return indices

    ###########################################################################
    # function and argument handling

    def _parse_function(self, ins, token):
        """Parse a function starting with the given token."""
        ins.read(len(token))
        if token in self._simple:
            # apply functions
            parse_args = self._simple[token]
        else:
            fndict = self._complex[token]
            presign = ins.skip_blank_read_if(fndict)
            if presign:
                token += presign
            try:
                parse_args = fndict[presign]
            except KeyError:
                raise error.BASICError(error.STX)
        if token == tk.FN:
            fnname = ins.read_name()
            # must not be empty
            error.throw_if(not fnname, error.STX)
            # obtain function
            function = self.user_functions.get(fnname)
            # get syntax
            parse_args = partial(self._gen_parse_arguments, length=function.number_arguments())
            fn = function.evaluate
        else:
            fn = self._callbacks[token]
        return fn(parse_args(ins))

    ###########################################################################
    # argument generators

    def _no_argument(self, ins):
        """No arguments to parse."""
        return
        yield # pragma: no cover

    def _gen_parse_arguments(self, ins, length=1):
        """Parse a comma-separated list of arguments."""
        if not length:
            return
        ins.require_read((b'(',))
        for i in range(length-1):
            yield self.parse(ins)
            ins.require_read((b',',))
        yield self.parse(ins)
        ins.require_read((b')',))

    def _gen_parse_arguments_optional(self, ins, length):
        """Parse a comma-separated list of arguments, last one optional."""
        ins.require_read((b'(',))
        yield self.parse(ins)
        for _ in range(length-2):
            ins.require_read((b',',))
            yield self.parse(ins)
        if ins.skip_blank_read_if((b',',)):
            yield self.parse(ins)
        else:
            yield None
        ins.require_read((b')',))

    def _gen_parse_one_optional_argument(self, ins):
        """Parse a single, optional argument."""
        if ins.skip_blank_read_if((b'(',)):
            yield self.parse(ins)
            ins.require_read((b')',))
        else:
            yield None

    def _gen_parse_call_extension(self, ins):
        """Parse an extension function."""
        yield ins.read_name()
        if ins.skip_blank_read_if((b'(',)):
            while True:
                yield self.parse(ins)
                if not ins.skip_blank_read_if((b',',)):
                    break
            ins.require_read((b')',))
        else:
            yield None

    ###########################################################################
    # special cases

    def _gen_parse_ioctl(self, ins):
        """Parse IOCTL$ syntax."""
        ins.require_read((b'(',))
        ins.skip_blank_read_if((b'#',))
        yield self.parse(ins)
        ins.require_read((b')',))

    def _gen_parse_instr(self, ins):
        """Parse INSTR syntax."""
        ins.require_read((b'(',))
        # followed by comma so empty will raise STX
        s = self.parse(ins)
        yield s
        if isinstance(s, values.Number):
            ins.require_read((b',',))
            yield self.parse(ins)
        ins.require_read((b',',))
        yield self.parse(ins)
        ins.require_read((b')',))

    def _gen_parse_input(self, ins):
        """Parse INPUT$ syntax."""
        ins.require_read((b'(',))
        yield self.parse(ins)
        if ins.skip_blank_read_if((b',',)):
            ins.skip_blank_read_if((b'#',))
            yield self.parse(ins)
        else:
            yield None
        ins.require_read((b')',))

    def _gen_parse_varptr_str(self, ins):
        """Parse VARPTR$ syntax."""
        ins.require_read((b'(',))
        yield ins.read_name()
        yield self.parse_indices(ins)
        ins.require_read((b')',))

    def _gen_parse_varptr(self, ins):
        """Parse VARPTR syntax."""
        ins.require_read((b'(',))
        if ins.skip_blank_read_if((b'#',)):
            yield self.parse(ins)
        else:
            yield ins.read_name()
            yield self.parse_indices(ins)
        ins.require_read((b')',))



"""
PC-BASIC - converter package
File format conversions and lexer

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .tokeniser import *
from .lister import *
from .protect import *



"""
PC-BASIC - audio_beep.py
Sound implementation through the linux beep utility

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import threading
import subprocess
from collections import deque
import time
import sys

from ..compat import WIN32, which

if WIN32:
    import winsound  # pylint: disable=import-error
    fcntl = None
else:
    import fcntl  # pylint: disable=import-error
    winsound = None

from .audio import AudioPlugin
from .base import audio_plugins, InitFailed


@audio_plugins.register('beep')
class AudioBeep(AudioPlugin):
    """Audio plugin based on the PC speaker."""

    def __init__(self, audio_queue, **kwargs):
        """Initialise sound system."""
        if WIN32:
            self.beeper = WinBeeper
        else:
            if LinuxBeeper.ok():
                self.beeper = LinuxBeeper
            else:
                self.beeper = Beeper
        if not self.beeper.ok():
            raise InitFailed('Beeper not supported')
        # sound generators for each voice
        self.generators = [deque(), deque(), deque(), deque()]
        AudioPlugin.__init__(self, audio_queue)

    def tone(self, voice, frequency, duration, loop, volume):
        """Enqueue a tone."""
        if voice == 0:
            self.generators[voice].append(self.beeper(
                    frequency, duration, loop, volume))

    def hush(self):
        """Stop sound."""
        for voice in range(4):
            self._next_tone[voice] = None
            while self.generators[voice]:
                self.generators[voice].popleft()
        self.beeper.hush()

    def _work(self):
        """Replenish sample buffer."""
        for voice in range(4):
            if self._next_tone[voice] is None or self._next_tone[voice].loop:
                try:
                    self._next_tone[voice] = self.generators[voice].popleft()
                except IndexError:
                    if self._next_tone[voice] is None:
                        continue
            self._next_tone[voice] = self._next_tone[voice].emit()


class Beeper(object):
    """Manage external beeper."""

    def __init__(self, frequency, duration, loop, dummy_volume):
        """Initialise beeper."""
        self._frequency = frequency
        self._duration = duration
        self._proc = None
        self.loop = loop

    @staticmethod
    def ok():
        """This beeper is supported."""
        return which('beep')

    @staticmethod
    def hush():
        """Stop sound."""
        subprocess.call('beep -f 1 -l 0'.split())

    def emit(self):
        """Emit a sound."""
        if not self._proc or (self.loop and self._proc.poll() is not None):
            if self._frequency == 0 or self._frequency == 32767:
                self._proc = subprocess.Popen(
                    'sleep {0}'.format(self._duration).split())
            else:
                self._proc = subprocess.Popen(
                    'beep -f {freq} -l {dur}'.format(
                        freq=self._frequency, dur=self._duration*1000,
                    ).split())
        # return self if still busy, None otherwise
        if self._proc and self._proc.poll() is None:
            return self
        else:
            return None


class WinBeeper(Beeper):
    """Manage speaker beeps through winsound."""

    @staticmethod
    def ok():
        """This beeper is supported when winsound is available."""
        return winsound is not None

    @staticmethod
    def hush():
        """Stop sound."""

    def emit(self):
        """Emit a sound."""
        if not self._proc or (self.loop and not self._proc.is_alive()):
            self._proc = threading.Thread(
                target=self._beep, args=(self._frequency, self._duration, self.loop))
            self._proc.start()
        # return self if still busy, None otherwise
        if self._proc and self._proc.is_alive():
            return self
        else:
            return None

    @staticmethod
    def _beep(frequency, duration, loop):
        """Beeping thread target."""
        if frequency < 37 or frequency >= 32767:
            time.sleep(duration)
        else:
            winsound.Beep(int(frequency), int(duration*1000))


KIOCSOUND = 0x4B2F
CLOCK_TICK_RATE = 1193180

class LinuxBeeper(WinBeeper):
    """Manage speaker beeps through Linux ioctl."""

    @staticmethod
    def ok():
        """This beeper is supported if the ioctl call works."""
        try:
            fcntl.ioctl(sys.stdout, KIOCSOUND, 0)
        except EnvironmentError:
            return False
        return True

    @staticmethod
    def hush():
        """Stop sound."""
        fcntl.ioctl(sys.stdout, KIOCSOUND, 0)

    @staticmethod
    def _beep(frequency, duration, loop):
        """Beeping thread target."""
        if frequency < 37 or frequency >= 32767:
            fcntl.ioctl(sys.stdout, KIOCSOUND, 0)
        else:
            fcntl.ioctl(sys.stdout, KIOCSOUND, int(CLOCK_TICK_RATE // frequency))
        time.sleep(duration)
        if not loop:
            fcntl.ioctl(sys.stdout, KIOCSOUND, 0)



"""
PC-BASIC - api.py
Session API

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import io

from ..compat import text_type

from .base import error
from .devices import NameWrapper
from . import implementation
from . import state

from ..data import read_codepage as codepage
from ..data import read_fonts as font
from .values import TYPE_TO_CLASS as SIGILS


class Session(object):
    """Public API to BASIC session."""

    def __init__(self, **kwargs):
        """Set up session object."""
        self._kwargs = kwargs
        self._impl = None

    def __enter__(self):
        """Context guard."""
        return self

    def __exit__(self, ex_type, ex_val, tb):
        """Context guard."""
        self.close()
        # catch Exit and Break events
        if ex_type in (error.Exit, error.Break):
            return True

    def __getstate__(self):
        """Pickle the session."""
        pickle_dict = self.__dict__.copy()
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Unpickle and resume the session."""
        self.__dict__.update(pickle_dict)

    def start(self):
        """Start the session."""
        if not self._impl:
            self._impl = implementation.Implementation(**self._kwargs)
            return True
        return False

    def attach(self, interface=None):
        """Attach interface to interpreter session."""
        self.start()
        self._impl.attach_interface(interface)
        return self

    def bind_file(self, file_name_or_object, name=None, create=False):
        """Bind a native file name or Python stream to a BASIC file name."""
        self.start()
        # if a file name, resolve
        if (
                not isinstance(file_name_or_object, (bytes, text_type))
                or os.path.isfile(file_name_or_object)
            ):
            # if it's an object or the file name exists, use it
            return self._impl.files.get_device(b'@:').bind(file_name_or_object, name)
        elif create and (
                not os.path.dirname(file_name_or_object) or
                os.path.isdir(os.path.dirname(file_name_or_object))
            ):
            # if it doesn't and we're allowed to create and the directory exists, create new
            return self._impl.files.get_device(b'@:').bind(file_name_or_object, name)
        # not resolved, try to use/create as internal name
        return NameWrapper(self._impl.codepage, file_name_or_object)

    def execute(self, command, as_type=None):
        """Execute a BASIC statement."""
        self.start()
        if as_type is None:
            as_type = type(command)
        output = io.BytesIO() if as_type == bytes else io.StringIO()
        with self._impl.io_streams.activate():
            self.add_pipes(output_streams=output)
            for cmd in command.splitlines():
                if isinstance(cmd, text_type):
                    cmd = self._impl.codepage.unicode_to_bytes(cmd)
                self._impl.execute(cmd)
            self.remove_pipes(output_streams=output)
        return output.getvalue()

    def add_pipes(self, input_streams=None, output_streams=None):
        """Add input/output pipes to session."""
        self.start()
        self._impl.io_streams.add_pipes(input_streams, output_streams)

    def remove_pipes(self, input_streams=None, output_streams=None):
        """Remove input/output pipes from session."""
        self.start()
        self._impl.io_streams.remove_pipes(input_streams, output_streams)

    def evaluate(self, expression):
        """Evaluate a BASIC expression."""
        self.start()
        with self._impl.io_streams.activate():
            if isinstance(expression, text_type):
                expression = self._impl.codepage.unicode_to_bytes(expression)
            return self._impl.evaluate(expression)

    def set_variable(self, name, value):
        """Set a variable in memory."""
        self.start()
        if isinstance(name, text_type):
            name = name.encode('ascii')
        name = name.upper()
        if name.split(b'(')[0][-1:] not in SIGILS:
            raise ValueError('Sigil must be explicit')
        self._impl.set_variable(name, value)

    def get_variable(self, name, as_type=None):
        """Get a variable in memory."""
        self.start()
        if isinstance(name, text_type):
            name = name.encode('ascii')
        if name.split(b'(')[0][-1:] not in SIGILS:
            raise ValueError('Sigil must be explicit')
        return self._impl.get_variable(name, as_type)

    def convert(self, value, to_type):
        """Convert a Python value to another type, consistent with BASIC rules."""
        self.start()
        return self._impl.get_converter(type(value), to_type)(value)

    def press_keys(self, keys):
        """Insert keypresses."""
        self.start()
        self._impl.keyboard.inject_keystrokes(keys)

    def get_chars(self, as_type=bytes):
        """Get currently displayed characters, as tuple of list of bytes / unicode."""
        self.start()
        return self._impl.text_screen.get_chars(as_type=as_type)

    def get_pixels(self):
        """Get currently displayed pixels, as tuple of tuples of int attributes."""
        self.start()
        return self._impl.display.vpage.pixels[:, :].to_rows()

    def greet(self):
        """Emit the interpreter greeting and show the key bar."""
        self.start()
        self._impl.execute(implementation.GREETING)

    def interact(self):
        """Interactive interpreter session."""
        self.start()
        with self._impl.io_streams.activate():
            self._impl.interact()

    def suspend(self, session_filename):
        """Save session object to file."""
        state.save_session(self, session_filename)

    @classmethod
    def resume(self, session_filename):
        """Load new session object from file."""
        return state.load_session(session_filename)

    def close(self):
        """Close the session."""
        if self._impl:
            self._impl.close()

    @property
    def info(self):
        """Get a session information object."""
        self.start()
        return SessionInfo(self)

    def set_hook(self, step_function):
        """Set function to be called on interpreter step."""
        self.start()
        self._impl.interpreter.step = step_function


class SessionInfo(object):
    """Retrieve information about current session."""

    def __init__(self, session):
        """Initialise the SessionInfo object."""
        self._session = session
        self._impl = session._impl

    def repr_scalars(self):
        """Get a representation of all scalars."""
        return repr(self._impl.scalars)

    def repr_arrays(self):
        """Get a representation of all arrays."""
        return repr(self._impl.scalars)

    def repr_strings(self):
        """Get a representation of string space."""
        return repr(self._impl.strings)

    def repr_text_screen(self):
        """Get a representation of the text screen."""
        return repr(self._impl.display.text_screen)

    def repr_program(self):
        """Get a marked-up hex dump of the program."""
        return repr(self._impl.program)

    def get_current_code(self, as_type=bytes):
        """Obtain statement being executed."""
        if self._impl.interpreter.run_mode:
            codestream = self._impl.program.bytecode
            bytepos = codestream.tell()
            from_line = self._impl.program.get_line_number(bytepos-1)
            try:
                codestream.seek(self._impl.program.line_numbers[from_line]+1)
                _, output, _ = self._impl.lister.detokenise_line(codestream)
                code_line = bytes(output)
            except KeyError:
                code_line = b''
        else:
            codestream = self._impl.interpreter.direct_line
            bytepos = codestream.tell()
            codestream.seek(0)
            code_line = bytes(
                self._impl.lister.detokenise_compound_statement(codestream)[0]
            )
        codestream.seek(bytepos)
        return self._session.convert(code_line, as_type)



"""
PC-BASIC - interface.interface
Interface class

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import sys
import threading
import logging
import traceback

from ..compat import queue

from ..basic.base import signals
from .base import InitFailed, video_plugins, audio_plugins, WAIT_MESSAGE
from .audio import AudioPlugin


# millisecond delay
DELAY = 12


class Interface(object):
    """User interface for PC-BASIC session."""

    def __init__(self, try_interfaces=(), audio_override=None, wait=False, **kwargs):
        """Initialise interface."""
        self._input_queue = queue.Queue()
        self._video_queue = queue.Queue()
        self._audio_queue = queue.Queue()
        self._wait = wait
        self._video, self._audio = None, None
        for video in try_interfaces:
            try:
                self._video = video_plugins[video](self._input_queue, self._video_queue, **kwargs)
            except KeyError:
                logging.error('Unknown video plugin `%s`', video)
            except InitFailed as e:
                logging.info('Could not initialise video plugin `%s`: %s', video, e)
            if self._video:
                break
        else:
            # video plugin is necessary, fail without it
            raise InitFailed('Failed to initialise any video plugin.')
        audio = audio_override or video
        try:
            self._audio = audio_plugins[audio](self._audio_queue, **kwargs)
        except KeyError:
            # ignore if an interface has no audio, but not if an override doesn't exist
            if audio_override and audio_override not in ('none', 'false'):
                logging.error('Unknown audio plugin `%s`', audio)
        except InitFailed as e:
            logging.info('Could not initialise audio plugin `%s`: %s', audio, e)
        if not self._audio:
            # audio fallback to no-plugin
            self._audio = AudioPlugin(self._audio_queue, **kwargs)

    def get_queues(self):
        """Retrieve interface queues."""
        return self._input_queue, self._video_queue, self._audio_queue

    def launch(self, target, *args, **kwargs):
        """Start an interactive interpreter session."""
        thread = threading.Thread(target=self._thread_runner, args=(target,) + args, kwargs=kwargs)
        try:
            # launch the BASIC thread
            thread.start()
            # run the interface
            self.run()
        except Exception as e:
            logging.error('Fatal error in interface')
            logging.error(''.join(traceback.format_exception(*sys.exc_info())))
        finally:
            self.quit_input()
            thread.join()

    def _thread_runner(self, target, *args, **kwargs):
        """Session runner."""
        try:
            target(*args, **kwargs)
        finally:
            if self._wait:
                self.pause(WAIT_MESSAGE)
            self.quit_output()

    def run(self):
        """Start the main interface event loop."""
        with self._audio:
            with self._video:
                while self._audio.alive or self._video.alive:
                    # ensure both queues are drained
                    self._video.cycle()
                    self._audio.cycle()
                    if not self._audio.busy and not self._video.busy:
                        # nothing to do, come back later
                        self._video.sleep(DELAY)

    def pause(self, message):
        """Pause and wait for a key."""
        self._video_queue.put(signals.Event(signals.VIDEO_SET_CAPTION, (message,)))
        self._video_queue.put(signals.Event(signals.VIDEO_SHOW_CURSOR, (False, False)))
        while True:
            signal = self._input_queue.get()
            if signal.event_type in (signals.KEYB_DOWN, signals.QUIT):
                self._video_queue.put(signals.Event(signals.VIDEO_SET_CAPTION, (u'',)))
                return signal

    def quit_input(self):
        """Send signal through the input queue to quit BASIC."""
        self._input_queue.put(signals.Event(signals.QUIT))
        # drain video queue (joined in other thread)
        while not self._video_queue.empty():
            try:
                signal = self._video_queue.get(False)
            except queue.Empty:
                continue
            self._video_queue.task_done()
        # drain audio queue
        while not self._audio_queue.empty():
            try:
                signal = self._audio_queue.get(False)
            except queue.Empty:
                continue
            self._audio_queue.task_done()

    def quit_output(self):
        """Send signal through the output queues to quit plugins."""
        self._video_queue.put(signals.Event(signals.QUIT))
        self._audio_queue.put(signals.Event(signals.QUIT))



"""
PC-BASIC - display.colours
Palettes, colours and attributes

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..base import error
from ..base import signals


# 2-colour CGA (SCREEN 2) palette
CGA2_PALETTE = (0, 15)
# 16 colour CGA palette (CGA text mode)
CGA16_PALETTE = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15)
# EGA palette - these are the same colours as 16-colour CGA but picked from different coloursets
EGA_PALETTE = (0, 1, 2, 3, 4, 5, 20, 7, 56, 57, 58, 59, 60, 61, 62, 63)

# CGA 4-colour palettes
# palette 0: Black, Green, Red, Brown/Yellow
# palette 1: Black, Ugh, Yuck, Bleah
# "Mode 5" (SCREEN 1 + colorburst on RGB) has red instead of magenta
# low intensity
CGA4_LO_PALETTE_0 = (0, 2, 4, 6)
CGA4_LO_PALETTE_1 = (0, 3, 5, 7)
CGA4_LO_PALETTE_RED = (0, 3, 4, 7)
# high intensity
CGA4_HI_PALETTE_0 = (0, 10, 12, 14)
CGA4_HI_PALETTE_1 = (0, 11, 13, 15)
CGA4_HI_PALETTE_RED = (0, 11, 12, 15)

# PCjr/Tandy 4-colour palettes
# like CGA 4-colour, but low-intensity colours with high-intensity white
TANDY4_PALETTE_0 = (0, 2, 4, 6)
TANDY4_PALETTE_1 = (0, 3, 5, 15)

# CGA colourset
# note 'CGA brown' in position 6
COLOURS16 = (
    (0x00, 0x00, 0x00), (0x00, 0x00, 0xaa), (0x00, 0xaa, 0x00), (0x00, 0xaa, 0xaa),
    (0xaa, 0x00, 0x00), (0xaa, 0x00, 0xaa), (0xaa, 0x55, 0x00), (0xaa, 0xaa, 0xaa),
    (0x55, 0x55, 0x55), (0x55, 0x55, 0xff), (0x55, 0xff, 0x55), (0x55, 0xff, 0xff),
    (0xff, 0x55, 0x55), (0xff, 0x55, 0xff), (0xff, 0xff, 0x55), (0xff, 0xff, 0xff)
)
# EGA colourset
COLOURS64 = (
    (0x00, 0x00, 0x00), (0x00, 0x00, 0xaa), (0x00, 0xaa, 0x00), (0x00, 0xaa, 0xaa),
    (0xaa, 0x00, 0x00), (0xaa, 0x00, 0xaa), (0xaa, 0xaa, 0x00), (0xaa, 0xaa, 0xaa),
    (0x00, 0x00, 0x55), (0x00, 0x00, 0xff), (0x00, 0xaa, 0x55), (0x00, 0xaa, 0xff),
    (0xaa, 0x00, 0x55), (0xaa, 0x00, 0xff), (0xaa, 0xaa, 0x55), (0xaa, 0xaa, 0xff),
    (0x00, 0x55, 0x00), (0x00, 0x55, 0xaa), (0x00, 0xff, 0x00), (0x00, 0xff, 0xaa),
    (0xaa, 0x55, 0x00), (0xaa, 0x55, 0xaa), (0xaa, 0xff, 0x00), (0xaa, 0xff, 0xaa),
    (0x00, 0x55, 0x55), (0x00, 0x55, 0xff), (0x00, 0xff, 0x55), (0x00, 0xff, 0xff),
    (0xaa, 0x55, 0x55), (0xaa, 0x55, 0xff), (0xaa, 0xff, 0x55), (0xaa, 0xff, 0xff),
    (0x55, 0x00, 0x00), (0x55, 0x00, 0xaa), (0x55, 0xaa, 0x00), (0x55, 0xaa, 0xaa),
    (0xff, 0x00, 0x00), (0xff, 0x00, 0xaa), (0xff, 0xaa, 0x00), (0xff, 0xaa, 0xaa),
    (0x55, 0x00, 0x55), (0x55, 0x00, 0xff), (0x55, 0xaa, 0x55), (0x55, 0xaa, 0xff),
    (0xff, 0x00, 0x55), (0xff, 0x00, 0xff), (0xff, 0xaa, 0x55), (0xff, 0xaa, 0xff),
    (0x55, 0x55, 0x00), (0x55, 0x55, 0xaa), (0x55, 0xff, 0x00), (0x55, 0xff, 0xaa),
    (0xff, 0x55, 0x00), (0xff, 0x55, 0xaa), (0xff, 0xff, 0x00), (0xff, 0xff, 0xaa),
    (0x55, 0x55, 0x55), (0x55, 0x55, 0xff), (0x55, 0xff, 0x55), (0x55, 0xff, 0xff),
    (0xff, 0x55, 0x55), (0xff, 0x55, 0xff), (0xff, 0xff, 0x55), (0xff, 0xff, 0xff)
)

# composite palettes
# see http://nerdlypleasures.blogspot.co.uk/2013_11_01_archive.html
COMPOSITE = {
    'cga': (
        (0x00, 0x00, 0x00), (0x00, 0x6a, 0x2c), (0x00, 0x39, 0xff), (0x00, 0x94, 0xff),
        (0xca, 0x00, 0x2c), (0x77, 0x77, 0x77), (0xff, 0x31, 0xff), (0xc0, 0x98, 0xff),
        (0x1a, 0x57, 0x00), (0x00, 0xd6, 0x00), (0x77, 0x77, 0x77), (0x00, 0xf4, 0xb8),
        (0xff, 0x57, 0x00), (0xb0, 0xdd, 0x00), (0xff, 0x7c, 0xb8), (0xff, 0xff, 0xff)
    ),
    'tandy': (
        (0x00, 0x00, 0x00), (0x7c, 0x30, 0x00), (0x00, 0x75, 0x00), (0x00, 0xbe, 0x00),
        (0x00, 0x47, 0xee), (0x77, 0x77, 0x77), (0x00, 0xbb, 0xc4), (0x00, 0xfb, 0x3f),
        (0xb2, 0x0f, 0x9d), (0xff, 0x1e, 0x0f), (0x77, 0x77, 0x77), (0xff, 0xb8, 0x00),
        (0xb2, 0x44, 0xff), (0xff, 0x78, 0xff), (0x4b, 0xba, 0xff), (0xff, 0xff, 0xff)
    ),
    'pcjr': (
        (0x00, 0x00, 0x00), (0x98, 0x20, 0xcb), (0x9f, 0x1c, 0x00), (0xff, 0x11, 0x71),
        (0x00, 0x76, 0x00), (0x77, 0x77, 0x77), (0x5b, 0xaa, 0x00), (0xff, 0xa5, 0x00),
        (0x00, 0x4e, 0xcb), (0x74, 0x53, 0xff), (0x77, 0x77, 0x77), (0xff, 0x79, 0xff),
        (0x00, 0xc8, 0x71), (0x00, 0xcc, 0xff), (0x00, 0xfa, 0x00), (0xff, 0xff, 0xff)
    ),
    # not used, retained for reference
    'cga_old': (
        (0x00, 0x00, 0x00), (0x00, 0x71, 0x00), (0x00, 0x3f, 0xff), (0x00, 0xab, 0xff),
        (0xc3, 0x00, 0x67), (0x73, 0x73, 0x73), (0xe6, 0x39, 0xff), (0x8c, 0xa8, 0xff),
        (0x53, 0x44, 0x00), (0x00, 0xcd, 0x00), (0x73, 0x73, 0x73), (0x00, 0xfc, 0x7e),
        (0xff, 0x39, 0x00), (0xe2, 0xca, 0x00), (0xff, 0x7c, 0xf4), (0xff, 0xff, 0xff)
    ),
}


# CGA mono intensities
INTENSITY16 = range(0x00, 0x100, 0x11)
# MDA text intensities: black, dark green, green, bright green
INTENSITY_MDA_MONO = (0x00, 0x40, 0xc0, 0xff)
# SCREEN 10 intensities
INTENSITY_EGA_MONO = (0x00, 0xaa, 0xff)

# monochrome EGA, these refer to the fixed pseudocolor palette defined in EGAMonoMode
# from GW-BASIC manual:
# Attribute Value	Displayed Pseudo-Color
# 0	Off
# 1	On, normal intensity
# 2	Blink
# 3	On, high intensity
# however, IBM EGA manual suggests these 4 are mapped to 16 *character* attributes differently:
# black, normal, black, normal, blink, bright, blink, bright, black, normal, black, normal, blink, bright, blink, bright
# it also doesn't mention being able to set the palette to the 9 pseudocolours of the GW manual
# but since we only have 2bpp I think graphics COLOR will use values 0-3, not 0-15
EGA_MONO_PALETTE = (0, 4, 1, 8)

# 64k EGA - we have 4 colours simultaneously out of a palette of 16
# from IBM's EGA manual they are black, blue, red, white
# however they are mapped to 16 character attributes
# black, blue, black, blue, red, white, red, white, black, blue, black, blue, red, white, red, white
EGA_64K_PALETTE = (0, 1, 4, 15)


#######################################################################################
# tinted monochrome monitors

MONO_TINT = {
    'green': (0, 255, 0),
    'amber': (255, 128, 0),
    'grey': (255, 255, 255),
    'mono': (0, 255, 0),
}

# apparent intensity
# these correspond to NTSC, ITU-R Recommendation BT.  601-2
# see https://poynton.ca/notes/colour_and_gamma/ColorFAQ.html#RTFToC3
#_RGB_INTENSITY = (0.299, 0.587, 0.114)
# some monitors seemed to show increasing intensity for the 16 shades,
# see https://www.vogons.org/viewtopic.php?t=29101
# also https://nerdlypleasures.blogspot.com/2014/03/the-monochrome-experience-cga-ega-and.html
_RGB_INTENSITY = (0.5, 0.3, 0.2)


def _adjust_tint(rgb, mono_tint, mono):
    """Convert (r, g, b) tuple to tinted monochrome."""
    if not mono:
        return rgb
    intensity = sum(_value * _weight for _value, _weight in zip(rgb, _RGB_INTENSITY))
    return tuple(int(_tint * intensity) // 255 for _tint in mono_tint)




###############################################################################
# colour mappers

class _ColourMapper(object):
    """Palette and colourset."""

    # override these

    # palette - lookup table that maps the valid attributes to colour values
    # these are "palette attributes" - e.g. the 16 foreground attributes for text mode.
    default_palette = ()
    num_attr = 0

    # colour set - maps the valid colour values to RGB
    # int values that can be used as the right hand side of a palette assignment
    _colours = ()

    # interpret colours as monochrome intensity
    _mono_tint = MONO_TINT['mono']
    _mono = False

    def __init__(self, queues, adapter, monitor, colorswitch):
        """Initialise colour map."""
        self._queues = queues
        if monitor in MONO_TINT:
            self._mono = monitor in MONO_TINT
            self._mono_tint = MONO_TINT[monitor]
        self.set_colorswitch(colorswitch)
        self._palette = list(self.default_palette)
        # policy on PALETTE changes
        # effective palette change is an error in CGA
        if adapter in ('cga', 'cga_old', 'mda', 'hercules', 'olivetti'):
            self._palette_change_policy = 'error'
        # ignore palette changes in Tandy/PCjr SCREEN 0
        elif adapter in ('tandy', 'pcjr') and self.num_attr == 256:
            self._palette_change_policy = 'deny'
        else:
            self._palette_change_policy = 'allow'
        self.submit()

    def allows_palette_change(self):
        """Check if the video mode allows palette change."""
        if self._palette_change_policy == 'error':
            raise error.BASICError(error.IFC)
        return self._palette_change_policy != 'deny'

    def reset(self):
        """Reset to default palette."""
        self._palette = list(self.default_palette)
        self.submit()

    @property
    def num_palette(self):
        """Number of values in palette."""
        return len(self.default_palette)

    @property
    def num_colours(self):
        """Number of values in colour set."""
        return len(self._colours)

    def split_attr(self, attr):
        """Split attribute byte into constituent parts."""
        return attr & 0xf, 0, False, False

    def _get_rgb_table(self):
        """List of RGB/blink/underline for all attributes, given a palette."""
        return (
            [self.attr_to_rgb(_attr) for _attr in range(self.num_attr)],
            None
        )

    def attr_to_rgb(self, attr):
        """Convert attribute to RGB/blink/underline, given a palette."""
        rgb = _adjust_tint(self._colours[self._palette[attr]], self._mono_tint, self._mono)
        return rgb, rgb, False, False

    def get_cga4_palette(self):
        """CGA palette setting (accessible from memory)."""
        return 1

    def set_cga4_palette(self, num):
        """Set the default 4-colour CGA palette."""

    def set_cga4_intensity(self, high):
        """Set/unset the palette intensity."""

    def set_colorswitch(self, colorswitch):
        """Set the SCREEN colorswitch parameter."""
        # in SCREEN 0, set the colorburst
        # in SCREEN 1, set the colorburst (inverted)
        # in SCREEN 2, ignore parameter and turn off colorburst

    def set_colorburst(self, on):
        """Set the NTSC colorburst bit."""
        # not colourburst capable
        return False

    # palette methods

    def set_all(self, new_palette, force=False):
        """Set the colours for all attributes."""
        if force or self.allows_palette_change():
            self._palette = list(new_palette)
            self.submit()

    def set_entry(self, index, colour, force=False):
        """Set a new colour for a given attribute."""
        if force or self.allows_palette_change():
            self._palette[index] = colour
            # in text mode, we'd be setting more than one attribute
            # e.g. all true attributes with this number as foreground or background
            # and attr_to_rgb decides which
            self.submit()

    def get_entry(self, index):
        """Retrieve the colour for a given attribute."""
        return self._palette[index]

    def submit(self):
        """Submit to interface."""
        # all attributes split into foreground RGB, background RGB, blink and underline
        rgb_table, compo_parms = self._get_rgb_table()
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_PALETTE, (rgb_table, compo_parms)
        ))

    # COLOR statement helpers

    def color(self, current_attr, fore, back, bord):
        """Mode-dependent helper function for COLOR statement."""
        if fore is None:
            fore = current_attr
        if back is None:
            # graphics mode bg is always 0; sets palette instead
            back = self.get_entry(0)
        # for screens other than 1, no distinction between 3rd parm zero and not supplied
        bord = bord or 0
        error.range_check(0, 255, bord)
        max_attr = self.num_attr - 1
        max_colour = self.num_colours - 1
        error.range_check(1, max_attr, fore)
        error.range_check(0, max_colour, back)
        # in screen 7 and 8, only low intensity palette is used for background
        self.set_entry(0, back % 8, force=True)
        # set attr, don't change border
        return fore, None

    # memory access

    def get_colour_info_byte(self):
        """Helper for PEEK(1126)."""
        return -1


class _TextColourMixin(object):
    """Translate text attributes to palette attributes."""

    num_attr = 256

    def split_attr(self, attr):
        """Split textmode attribute byte into constituent parts."""
        # 7  6 5 4  3 2 1 0
        # Bl b b b  f f f f
        back = (attr >> 4) & 7
        blink = (attr >> 7) == 1
        fore = attr & 0xf
        underline = False
        return fore, back, blink, underline

    def attr_to_rgb(self, attr):
        """Convert attribute to RGB/blink/underline, given a palette."""
        fore, back, blink, underline = self.split_attr(attr)
        fore_rgb = _adjust_tint(self._colours[self._palette[fore]], self._mono_tint, self._mono)
        back_rgb = _adjust_tint(self._colours[self._palette[back]], self._mono_tint, self._mono)
        return fore_rgb, back_rgb, blink, underline

    def set_colorswitch(self, colorswitch):
        """Set the SCREEN colorswitch parameter."""
        self.set_colorburst(colorswitch)

    def color(self, current_attr, fore, back, bord):
        """Helper function for COLOR in text mode (SCREEN 0)."""
        if fore is not None:
            # allow twice the number of foreground attributes (16) - because of blink
            error.range_check(0, self.num_palette*2-1, fore)
            # COLOR > 17 means blink, but the blink bit is the top bit of the true attribute
            bit, fore = divmod(fore, self.num_palette)
        else:
            # use explicit fore-back-bit split here, we don't care if it's actually blink/underline
            # and split_attr/join_attr are being overridden
            fore = current_attr & 0xf
            bit = (current_attr >> 7) == 1
        if back is not None:
            # allow background attributes up to 15 though highest bit is ignored
            error.range_check(0, self.num_palette-1, back)
        else:
            back = (current_attr >> 4) & 7
        if bord is not None:
            error.range_check(0, self.num_palette-1, bord)
        # set attribute and border
        attr = ((bit & 1) << 7) + ((back & 7) << 4) + (fore & 0xf)
        return attr, bord


class HerculesColourMapper(_ColourMapper):
    """Hercules 16-greyscale palette."""

    # Hercules graphics has no PALETTE
    # see MS KB 21839, https://jeffpar.github.io/kbarchive/kb/021/Q21839/
    default_palette = (0, 1)
    num_attr = 2
    _colours = ((0, 0, 0), (255, 255, 255))
    _mono = True

    def color(self, current_attr, fore, back, bord):
        """Mode-dependent helper function for COLOR statement."""
        raise error.BASICError(error.IFC)


class _CGAColourMapper(_ColourMapper):
    """CGA 2-colour, 16-colour palettes."""

    _colours = COLOURS16

    def __init__(self, queues, adapter, monitor, colorswitch):
        """Initialise colour map."""
        self._has_colorburst = adapter in ('cga', 'cga_old', 'pcjr', 'tandy')
        # monochrome
        self._force_mono = monitor in MONO_TINT
        # rgb monitor
        self._force_colour = not self._force_mono and monitor != 'composite'
        _ColourMapper.__init__(self, queues, adapter, monitor, colorswitch)

    def set_colorburst(self, colour_on):
        """Set the NTSC colorburst bit."""
        # On a composite monitor with CGA adapter (not EGA, VGA):
        # - on SCREEN 2 this enables artifacting
        # - on SCREEN 1 and 0 this switches between colour and greyscale
        # On an RGB monitor:
        # - on SCREEN 1 this switches between mode 4/5 palettes (RGB)
        # - ignored on other screens
        if self._has_colorburst:
            self._mono = self._force_mono or (not colour_on and not self._force_colour)
        # reset the palette to reflect the new mono situation, submit to interface
        self.reset()


class _CompositeMixin(object):
    """Overrides to deal with NTSC composite artifacts."""

    def __init__(self, adapter, monitor):
        """Initialise colour map."""
        self._adapter = adapter
        self._has_composite = (
            monitor == 'composite' and adapter in COMPOSITE
        )
        self._composite = False

    def set_colorburst(self, colour_on):
        """Set the NTSC colorburst bit."""
        # On a composite monitor with CGA adapter (not EGA, VGA):
        # - on SCREEN 2 this enables artifacting
        self._composite = colour_on and self._has_composite
        # reset the palette for artifactig, submit to interface
        self.reset()

    def _get_rgb_table(self):
        """List of RGB/blink/underline for all attributes, given a palette."""
        if self._composite:
            compo_palette = tuple((_c, _c, False, False) for _c in COMPOSITE[self._adapter])
            # 4bpp composite (16 shades), 1bpp original
            return (compo_palette, (4, 1))
        return _CGAColourMapper._get_rgb_table(self)


class CGA2ColourMapper(_CompositeMixin, _CGAColourMapper):
    """CGA 2-colour palettes, with composite support."""

    default_palette = CGA2_PALETTE
    num_attr = 2

    def __init__(self, queues, adapter, monitor, colorswitch):
        """Initialise colour map."""
        self._has_colorburst = adapter in ('cga', 'cga_old', 'pcjr', 'tandy')
        _CompositeMixin.__init__(self, adapter, monitor)
        _CGAColourMapper.__init__(self, queues, adapter, monitor, colorswitch)

    def set_colorswitch(self, colorswitch):
        """Set the SCREEN colorswitch parameter."""
        # in SCREEN 2 the colorswitch parameter has no effect
        # the color burst can only be changed through the CGA mode control register at port 03D8h
        self.set_colorburst(False)

    def color(self, current_attr, arg0, arg1, arg2):
        """Helper function for COLOR statement in SCREEN 2."""
        raise error.BASICError(error.IFC)


class Olivetti2ColourMapper(_CGAColourMapper):
    """Olivetti 2-colour palettes."""

    default_palette = CGA2_PALETTE
    num_attr = 2

    def color(self, current_attr, fore, arg1, arg2):
        """Helper function for COLOR statement in SCREEN 100."""
        if fore is None:
            fore = current_attr
        # for screens other than 1, no distinction between 3rd parm zero and not supplied
        arg2 = arg2 or 0
        error.range_check(0, 255, arg2)
        max_attr = self.num_attr - 1
        max_colour = self.num_colours - 1
        # Olivetti screen 4 2-colour can change foreground with COLOR but bg is always black
        error.range_check(0, max_colour, fore)
        if arg1:
            # mut be zero or unspecified
            raise error.BASICError(error.IFC)
        self.set_entry(1, fore, force=True)
        return None, None


class CGA16ColourMapper(_CGAColourMapper):
    """CGA 16-colour palettes."""

    default_palette = CGA16_PALETTE
    num_attr = 16


class CGA4ColourMapper(_CGAColourMapper):
    """CGA 4-colour palettes, with colorburst setting, intensity setting and mode 5."""

    num_attr = 4

    def __init__(self, queues, adapter, monitor, colorswitch):
        """Initialise colour map."""
        self._has_colorburst = adapter in ('cga', 'cga_old', 'pcjr', 'tandy')
        # pcjr/tandy does not have mode 5
        self._tandy = adapter in ('pcjr', 'tandy')
        self._has_mode_5 = adapter in ('cga', 'cga_old')
        self._low_intensity = False
        # start with the cyan-magenta-white palette
        self._palette_number = 1
        self._mode_5 = False
        _CGAColourMapper.__init__(self, queues, adapter, monitor, colorswitch)

    def get_cga4_palette(self):
        """CGA palette setting (accessible from memory)."""
        return self._palette_number

    def set_cga4_palette(self, num):
        """Set the default 4-colour CGA palette."""
        self._palette_number = num % 2
        self.reset()

    def set_cga4_intensity(self, high):
        """Set/unset the palette intensity."""
        self._low_intensity = not high
        self.reset()

    @property
    def mode_5(self):
        """Mode 5 is active."""
        return self._mode_5 and self._has_mode_5

    @property
    def default_palette(self):
        """Default palette."""
        if self._tandy:
            if self._palette_number:
                return TANDY4_PALETTE_1
            else:
                return TANDY4_PALETTE_0
        elif self._mode_5 and self._has_mode_5:
            if self._low_intensity:
                return CGA4_LO_PALETTE_RED
            else:
                return CGA4_HI_PALETTE_RED
        elif self._low_intensity:
            if self._palette_number:
                return CGA4_LO_PALETTE_1
            else:
                return CGA4_LO_PALETTE_0
        else:
            if self._palette_number:
                return CGA4_HI_PALETTE_1
            else:
                return CGA4_HI_PALETTE_0

    def set_colorswitch(self, colorswitch):
        """Set the SCREEN colorswitch parameter."""
        # in SCREEN 1 the SCREEN colorswitch parameter works the other way around
        self.set_colorburst(not colorswitch)

    def set_colorburst(self, colour_on):
        """Set the NTSC colorburst bit."""
        if not self._has_colorburst:
            return
        # On a composite monitor with CGA adapter (not EGA, VGA):
        # - on SCREEN 1 and 0 this switches between colour and greyscale
        # On an RGB monitor:
        # - on SCREEN 1 this switches between mode 4/5 palettes (RGB)
        if self._force_colour:
            # ega ignores colorburst; tandy and pcjr have no mode 5
            self._mode_5 = not colour_on
            self.set_cga4_palette(1)
        else:
            self._mono = self._force_mono or not colour_on
            # reset the palette for mono, submit to interface
            self.reset()

    def color(self, current_attr, back, pal, override):
        """Helper function for COLOR in SCREEN 1."""
        back = self.get_entry(0) if back is None else back
        if override is not None:
            # uses last entry as palette if given
            pal = override
        error.range_check(0, 255, back)
        if pal is not None:
            error.range_check(0, 255, pal)
            self.set_cga4_palette(pal % 2)
            palette = list(self.default_palette)
            palette[0] = back & 0xf
            self.set_all(palette, force=True)
        else:
            self.set_entry(0, back & 0xf, force=True)
        return None, None

    def get_colour_info_byte(self):
        """Helper for PEEK(1126)."""
        return self.get_entry(0) + 32 * self.get_cga4_palette()


class Tandy4ColourMapper(_ColourMapper):
    """Tandy 4@16-colour mapper (Tandy mode 4)."""

    _colours = COLOURS16
    # it's unclear to me what the default palette for this mode was
    default_palette = TANDY4_PALETTE_1
    num_attr = 4


class EGA4ColourMapper(_ColourMapper):
    """EGA 4@16-colour mapper (64k EGA)."""

    _colours = COLOURS16
    # it's unclear to me what the default palette for this mode was
    default_palette = EGA_64K_PALETTE
    num_attr = 4


class EGA16ColourMapper(_ColourMapper):
    """EGA 16@16-colour mapper."""

    _colours = COLOURS16
    default_palette = CGA16_PALETTE
    num_attr = 16


class EGA64ColourMapper(_ColourMapper):
    """EGA 16@64-colour mapper."""

    _colours = COLOURS64
    default_palette = EGA_PALETTE
    num_attr = 16

    def color(self, current_attr, fore, back, bord):
        """Helper function for COLOR statement in SCREEN 9."""
        if fore is None:
            fore = current_attr
        if back is None:
            # graphics mode bg is always 0; sets palette instead
            back = self.get_entry(0)
        # for screens other than 1, no distinction between 3rd parm zero and not supplied
        bord = bord or 0
        error.range_check(0, 255, bord)
        max_attr = self.num_attr - 1
        max_colour = self.num_colours - 1
        error.range_check(1, max_attr, fore)
        error.range_check(0, max_colour, back)
        # all background colours are accessible
        self.set_entry(0, back, force=True)
        # set attr, don't change border
        return fore, None



class EGA16TextColourMapper(_TextColourMixin, EGA16ColourMapper):
    """EGA 16@16-colour mapper for text."""


class EGA64TextColourMapper(_TextColourMixin, EGA64ColourMapper):
    """EGA 16@64-colour mapper for text."""

    # technically, VGA text does have underline
    # but it's set to an invisible scanline
    # so not, so long as we're not allowing to set the scanline


class MonoTextColourMapper(_TextColourMixin, _ColourMapper):
    """Attribute mapper for MDA-style text mode with underlining."""

    # herc attributes shld distinguish black, dim, normal, bright
    # see http://www.seasip.info/VintagePC/hercplus.html

    # MDA text attributes: http://www.seasip.info/VintagePC/mda.html
    # The attribute bytes mostly behave like a bitmap:
    #
    # Bit 1: Underline.
    # Bit 3: High intensity.
    # Bit 7: Blink
    # but there are eight exceptions:
    #
    # Attributes 00h, 08h, 80h and 88h display as black space.
    # Attribute 70h displays as black on green.
    # Attribute 78h displays as dark green on green.
    #       In fact, depending on timing and on the design of the monitor,
    #       it may have a bright green 'halo' where the dark green and bright green bits meet.
    # Attribute F0h displays as a blinking version of 70h
    #       (if blinking is enabled); as black on bright green otherwise.
    # Attribute F8h displays as a blinking version of 78h
    #       (if blinking is enabled); as dark green on bright green otherwise.

    # see also http://support.microsoft.com/KB/35148
    # --> archived on https://github.com/jeffpar/kbarchive/tree/master/kb/035/Q35148

    # https://www.pcjs.org/pubs/pc/reference/microsoft/kb/Q44412/

    # this should agree with palettes for EGA mono for COLOR values, see:
    # http://qbhlp.uebergeord.net/screen-statement-details-colors.html

    # EGA mono attributes are actually different on the intermediate, non-standard byte attributes
    # see https://nerdlypleasures.blogspot.com/2014/03/the-monochrome-experience-cga-ega-and.html
    # and http://www.vcfed.org/forum/showthread.php?50674-EGA-Monochrome-Compatibility

    _colours = tuple((_i, _i, _i) for _i in INTENSITY_MDA_MONO)

    @property
    def default_palette(self):
        """Default palette."""
        # this is actually ignored
        # remove after refactoring
        return (0,) * 16

    @property
    def num_palette(self):
        """Number of foreground attributes is the same as in colour text modes."""
        return 16

    def split_attr(self, attr):
        """Split attribute byte into constituent parts."""
        underline = (attr % 8) == 1
        blink = (attr & 0x80) != 0
        # background is almost always black
        back = 0
        # intensity set by bit 3
        fore = 2 if not (attr & 0x8) else 3
        # exceptions
        if attr in (0x00, 0x08, 0x80, 0x88):
            fore, back = 0, 0
        elif attr in (0x70, 0xf0):
            fore, back = 0, 2
        elif attr in (0x78, 0xf8):
            fore, back = 1, 3
        return fore, back, blink, underline

    def attr_to_rgb(self, attr):
        """Convert colour attribute to RGB/blink/underline, given a palette."""
        fore, back, blink, underline = self.split_attr(attr)
        # palette is ignored
        fore_rgb = _adjust_tint(self._colours[fore], self._mono_tint, self._mono)
        back_rgb = _adjust_tint(self._colours[back], self._mono_tint, self._mono)
        return fore_rgb, back_rgb, blink, underline


class EGAMonoColourMapper(_ColourMapper):
    """Colour mapper for EGA monochrome graphics mode (mode 10)."""

    # from GW-BASIC manual:
    # Color Value	Displayed Pseudo-Color
    # 0	Off
    # 1	Blink, off to on
    # 2	Blink, off to high intensity
    # 3	Blink, on to off
    # 4	On
    # 5	Blink, on to high intensity
    # 6	Blink, high intensity to off
    # 7	Blink, high intensity to on
    # 8	High intensity

    # fore, back intensities. blink is from fore to back
    _pseudocolours = (
        (0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)
    )

    default_palette = EGA_MONO_PALETTE
    num_attr = 4
    _colours = tuple((_i, _i, _i) for _i in INTENSITY_EGA_MONO)

    @property
    def num_colours(self):
        """Number of colour values."""
        return len(self._pseudocolours)

    def attr_to_rgb(self, attr):
        """Convert colour attribute to RGB/blink/underline, given a palette."""
        fore, back = self._pseudocolours[self._palette[attr] % len(self._pseudocolours)]
        # intensity 0, 1, 2 to RGB; apply mono tint
        fore_rgb = _adjust_tint(self._colours[fore], self._mono_tint, self._mono)
        back_rgb = _adjust_tint(self._colours[back], self._mono_tint, self._mono)
        # fore, back, blink, underline
        return fore_rgb, back_rgb, fore != back, False



"""
PC-BASIC - devices.diskfiles
Disk Files

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct
import ntpath
from contextlib import contextmanager

from ...compat import iteritems, iterchar

from ..base.bytestream import ByteStream
from ..base import error
from .devicebase import RawFile, TextFileBase, InputMixin, safe_io, TYPE_TO_MAGIC


# binary file interface: file interface +
#   seg
#   offset
#   length

# locking interface
#   number
#   lock()
#   unlock()


class BinaryFile(RawFile):
    """File class for binary (B, P, M) files on disk device."""

    def __init__(self, fhandle, filetype, number, mode, seg, offset, length, locks):
        """Initialise program file object and write header."""
        RawFile.__init__(self, fhandle, filetype, mode)
        # don't lock binary files
        # we need the Locks object to register file as open
        self._locks = locks
        self._number = number
        # binary file parameters
        self.seg, self.offset, self.length = 0, 0, 0
        if self.mode == b'O':
            self.write(TYPE_TO_MAGIC[filetype])
            if self.filetype == b'M':
                self.write(struct.pack(b'<HHH', seg, offset, length))
                self.seg, self.offset, self.length = seg, offset, length
        else:
            # drop magic byte
            self.read(1)
            if self.filetype == b'M':
                # self.length gets ignored: even the \x1a at the end is read
                header = self.read(6)
                if len(header) == 6:
                    self.seg, self.offset, self.length = struct.unpack('<HHH', header)
                else:
                    # truncated header
                    raise error.BASICError(error.BAD_FILE_MODE)

    def close(self):
        """Write EOF and close program file."""
        if self.mode == b'O':
            self.write(b'\x1a')
        RawFile.close(self)
        # no locking for binary files, but we do need to register it closed
        self._locks.close_file(self._number)


class TextFile(TextFileBase, InputMixin):
    """Text file on disk device."""

    def __init__(self, fhandle, filetype, number, mode, locks):
        """Initialise text file object."""
        TextFileBase.__init__(self, fhandle, filetype, mode)
        self._locks = locks
        self._number = number
        # in append mode, we need to start at end of file
        if self.mode == b'A':
            with safe_io():
                self._fhandle.seek(0, 2)

    def close(self):
        """Close text file."""
        if self.mode in (b'O', b'A'):
            # write EOF char
            with safe_io():
                self._fhandle.write(b'\x1a')
        TextFileBase.close(self)
        self._locks.close_file(self._number)

    def read(self, n):
        """Read num characters."""
        self._locks.try_access(self._number, b'R')
        return TextFileBase.read(self, n)

    def read_one(self):
        """Read one character, replacing CR LF with CR."""
        c = self.read(1)
        if not c:
            return c
        # report CRLF as CR
        # but LFCR, LFCRLF, LFCRLFCR etc pass unmodified
        if (c == b'\r' and self._previous != b'\n') and self.peek(1) == b'\n':
            last, char = self._previous, self._current
            self.read(1)
            self._previous, self._current = last, char
        return c

    def read_line(self):
        """Read line from text file, break on CR or CRLF (not LF)."""
        s = []
        while True:
            c = self.read_one()
            if not c or (c == b'\r' and self._previous != b'\n'):
                # break on CR, CRLF but allow LF, LFCR to pass
                break
            s.append(c)
            if len(s) == 255:
                c = b'\r' if self.peek(1) == b'\r' else None
                break
        return b''.join(s), c

    def write(self, s, can_break=True):
        """Write string to file."""
        self._locks.try_access(self._number, b'W')
        TextFileBase.write(self, s, can_break)

    def write_line(self, s=b''):
        """Write string and newline to file."""
        self.write(s + b'\r\n')

    def loc(self):
        """Get file pointer (LOC)."""
        with safe_io():
            if self.mode == b'I':
                tell = self._fhandle.tell() - len(self._readahead)
                return max(1, (127+tell) // 128)
            return self._fhandle.tell() // 128

    def lof(self):
        """Get length of file (LOF)."""
        with safe_io():
            current = self._fhandle.tell()
            self._fhandle.seek(0, 2)
            lof = self._fhandle.tell()
            self._fhandle.seek(current)
        return lof

    def lock(self, start, stop):
        """Lock the file."""
        # range bounds are ignored on text file
        # we need a tuple in case the other file checking the lock is a random file
        self._locks.acquire_record_lock(self._number, None, None)

    def unlock(self, start, stop):
        """Unlock the file."""
        self._locks.release_record_lock(self._number, None, None)


class FieldFile(TextFile):
    """Text file on FIELD."""

    def __init__(self, field, reclen):
        """Initialise text file object."""
        # don't let the field file use device locks
        TextFile.__init__(self, ByteStream(field.view_buffer()), b'D', None, b'I', Locks())
        self._field = field
        self._reclen = reclen

    @contextmanager
    def use_mode(self, mode):
        """Use in input or output mode."""
        self._switch_mode(mode)
        yield
        self._check_overflow()

    def __getstate__(self):
        """Pickle."""
        pickledict = self.__dict__
        pickledict['_pos'] = self._fhandle.tell()
        # can't pickle memoryview objects
        del pickledict['_fhandle']
        return pickledict

    def __setstate__(self, pickledict):
        """Unpickle."""
        pos = pickledict.pop('_pos')
        self. __dict__ = pickledict
        self._fhandle = ByteStream(self._field.view_buffer())
        self._fhandle.seek(pos)

    def _switch_mode(self, new_mode):
        """Switch to input or output mode and fix readahaed buffer."""
        if new_mode == b'I' and self.mode == b'O':
            self._fhandle.flush()
            self.mode = b'I'
        elif new_mode == b'O' and self.mode == b'I':
            self._fhandle.seek(-len(self._readahead), 1)
            self._readahead = []
            self._previous, self._current = b'', b''
            self.mode = b'O'

    def _check_overflow(self):
        """Check for FIELD OVERFLOW."""
        # FIELD overflow happens if last byte in record has been read or written
        if self._fhandle.tell() - len(self._readahead) >= self._reclen:
            raise error.BASICError(error.FIELD_OVERFLOW)

    def set_buffer(self, contents):
        """Set the contents of the buffer."""
        # take contents and pad with NULL to required size
        try:
            self._field.view_buffer()[:self._reclen] = contents.ljust(self._reclen, b'\0')
        except ValueError:
            # can't modify size of memoryview
            raise error.BASICError(error.FIELD_OVERFLOW)
        # reset field text file loc
        self._fhandle.seek(0)

    def get_buffer(self):
        """Get a copy of the contents of the buffer."""
        return bytearray(self._field.view_buffer()[:self._reclen])

    def write(self, bytestr, can_break=True):
        """Write bytes to buffer."""
        try:
            TextFile.write(self, bytestr, can_break)
        except ValueError:
            # can't modify size of memoryview
            raise error.BASICError(error.FIELD_OVERFLOW)


class RandomFile(RawFile):
    """Random-access file on disk device."""

    def __init__(self, fhandle, number, field, reclen, locks):
        """Initialise random-access file."""
        # note that for random files, output_stream must be a seekable stream.
        RawFile.__init__(self, fhandle, b'D', b'R')
        self.reclen = reclen
        self._locks = locks
        self._number = number
        # all text-file operations on a RANDOM file (PRINT, WRITE, INPUT, ...)
        # actually work on the FIELD buffer; the file stream itself is not
        # touched until PUT or GET.
        self._field_file = FieldFile(field, reclen)
        # position at start of file
        self._recpos = 0
        self._fhandle.seek(0)

    def close(self):
        """Close random-access file."""
        RawFile.close(self)
        self._locks.close_file(self._number)

    ##########################################################################
    # field text file operations

    def read(self, num):
        """Read a number of characters from the field buffer."""
        with self._field_file.use_mode(b'I'):
            return self._field_file.read(num)

    def input_entry(self, typechar, allow_past_end):
        """Read a number or string entry for INPUT """
        with self._field_file.use_mode(b'I'):
            # reading past end should give FIELD OVERFLOW, so suppress INPUT_PAST_END
            return self._field_file.input_entry(typechar, allow_past_end=True)

    def read_one(self):
        """Read a number of characters from the field buffer."""
        with self._field_file.use_mode(b'I'):
            return self._field_file.read_one()

    def read_line(self):
        """Read a line from the field buffer."""
        with self._field_file.use_mode(b'I'):
            return self._field_file.read_line()

    def write(self, s, can_break=True):
        """Write the string s to the field."""
        with self._field_file.use_mode(b'O'):
            self._field_file.write(s, can_break)

    def write_line(self, s=b''):
        """Write string and newline to the field buffer."""
        with self._field_file.use_mode(b'O'):
            self._field_file.write_line(s)

    @property
    def width(self):
        """Get the width setting on the field buffer."""
        return self._field_file.width

    def set_width(self, new_width):
        """Change the width setting on the field buffer."""
        self._field_file.set_width(new_width)

    @property
    def col(self):
        """Get the current column position on the field buffer."""
        return self._field_file.col

    ##########################################################################

    def eof(self):
        """Return whether we're past current end-of-file, for EOF."""
        return self._recpos * self.reclen > self.lof()

    def get(self, pos):
        """Read a record."""
        self._set_record_pos(pos)
        # exceptionally, GET is allowed if the file holding the lock is open for OUTPUT
        self._locks.try_record_access(self._number, self._recpos+1, self._recpos+1, b'R')
        if self.eof():
            contents = b'\0' * self.reclen
        else:
            with safe_io():
                contents = self._fhandle.read(self.reclen)
        # take contents and pad with NULL to required size
        self._field_file.set_buffer(contents)
        self._recpos += 1

    def put(self, pos):
        """Write a record."""
        self._set_record_pos(pos)
        self._locks.try_record_access(self._number, self._recpos+1, self._recpos+1, b'W')
        current_length = self.lof()
        with safe_io():
            if self._recpos > current_length:
                self._fhandle.seek(0, 2)
                numrecs = self._recpos - current_length
                self._fhandle.write(b'\0' * numrecs * self.reclen)
            self._fhandle.write(bytes(self._field_file.get_buffer()))
        self._recpos += 1

    def _set_record_pos(self, pos):
        """Move record pointer to new position."""
        if pos is not None:
            # first record is number 1
            with safe_io():
                self._fhandle.seek((pos-1) * self.reclen)
            self._recpos = pos - 1

    def loc(self):
        """Get number of record just past, for LOC."""
        return self._recpos

    def lof(self):
        """Get length of file, in bytes, for LOF."""
        with safe_io():
            current = self._fhandle.tell()
            self._fhandle.seek(0, 2)
            lof = self._fhandle.tell()
            self._fhandle.seek(current)
        return lof

    def lock(self, start, stop):
        """Lock range of records."""
        self._locks.acquire_record_lock(self._number, start, stop)

    def unlock(self, start, stop):
        """Unlock range of records."""
        self._locks.release_record_lock(self._number, start, stop)


###############################################################################
# Locks


class LockingParameters(object):
    """Record of a file's locking parameters."""

    def __init__(self, dos_name, mode, lock_type, access):
        """Build a record."""
        self.name = ntpath.basename(dos_name).upper()
        self.lock_set = set()
        self.lock_type = lock_type
        self.access = access
        self.mode = mode


class Locks(object):
    """Lock management."""

    def __init__(self):
        """Initialise locks."""
        # dict of LockingParameters objects, one for each open disk file, by file number
        self._locking_parameters = {}

    def list_open(self, name, exclude_number=None):
        """Retrieve a list of files open on the same disk device."""
        return [
            f for number, f in iteritems(self._locking_parameters)
            if f.name == ntpath.basename(name).upper() and number != exclude_number
        ]

    def open_file(self, name, number, mode, lock_type, access):
        """Register a disk file and try to acquire a file lock."""
        already_open = self.list_open(name)
        if mode in (b'O', b'A') and already_open:
            raise error.BASICError(error.FILE_ALREADY_OPEN)
        if not number:
            return
        for f in already_open:
            if (
                    # default mode: don't accept if SHARED/LOCK present
                    ((not lock_type) and f.lock_type) or
                    # LOCK READ WRITE: don't accept if already open
                    (lock_type == b'RW') or
                    # defined locking: don't accept if open in default mode
                    (lock_type and not f.lock_type) or
                    # LOCK READ or LOCK WRITE: accept based on ACCESS of open file
                    (
                        lock_type and lock_type != b'SHARED' and
                        f.access and set(iterchar(lock_type)) & set(iterchar(f.access))
                    ) or
                    (
                        f.lock_type and f.lock_type != b'SHARED' and
                        (
                            (access and set(iterchar(f.lock_type)) & set(iterchar(access))) or
                            # can't open with unspecified access if other is LOCK READ WRITE
                            (not access and set(iterchar(f.lock_type)) == {b'R', b'W'})
                        )
                    )
                ):
                raise error.BASICError(error.PERMISSION_DENIED)
        # setting this only after the lock acquisition makes the check asymmetric
        # which is what GW-BASIC does...
        # first file to open with unspecified access gets RW access
        # but second file gets checked for ''
        if lock_type and not access:
            access = b'RW'
        self._locking_parameters[number] = LockingParameters(name, mode, lock_type, access)

    def close_file(self, number):
        """Deregister disk file."""
        try:
            del self._locking_parameters[number]
        except KeyError:
            pass

    def try_access(self, number, access):
        """Attempt to access a file."""
        if not number:
            return
        this_file = self._locking_parameters[number]
        # access in violation of ACCESS declaration in OPEN: path/file access error
        if this_file.access and not (set(access) & set(this_file.access)):
            raise error.BASICError(error.PATH_FILE_ACCESS_ERROR)
        # access in violation of other's LOCK declation in OPEN: path/file access error
        others = self.list_open(this_file.name, number)
        for f in others:
            if (f.lock_type and f.lock_type != b'SHARED' and (set(f.lock_type) & set(access))):
                raise error.BASICError(error.PATH_FILE_ACCESS_ERROR)

    def try_record_access(self, number, start, stop, access=b'RW'):
        """Attempt to access a record."""
        self.try_access(number, access)
        self._try_record_lock(number, start, stop, allow_self=True, read_only=(access == b'R'))

    def _try_record_lock(self, number, start, stop, allow_self=True, read_only=False):
        """Attempt to access a record."""
        this_file = self._locking_parameters[number]
        other_locks = [
            f.lock_set for f in self.list_open(this_file.name, number if allow_self else None)
            # access parameter only exists to allow reading a record on locked OUTPUT file
            if not (f.mode in b'OA' and read_only)
        ]
        other_lock_set = set.union(*other_locks) if other_locks else set()
        # access in violation of other's LOCK#: permission denied
        # whole-file access sought
        if stop is None and start is None:
            if other_lock_set:
                raise error.BASICError(error.PERMISSION_DENIED)
        else:
            # range access sought
            for start_1, stop_1 in other_lock_set:
                if (
                        stop_1 is None and start_1 is None
                        or (start >= start_1 and start <= stop_1)
                        or (stop >= start_1 and stop <= stop_1)
                    ):
                    raise error.BASICError(error.PERMISSION_DENIED)

    def acquire_record_lock(self, number, start, stop):
        """Acquire a lock on a range of records."""
        self._try_record_lock(number, start, stop, allow_self=False)
        this_file = self._locking_parameters[number]
        this_file.lock_set.add((start, stop))

    def release_record_lock(self, number, start, stop):
        """Acquire a lock on a range of records."""
        this_file = self._locking_parameters[number]
        # permission denied if the exact record range wasn't given before
        try:
            this_file.lock_set.remove((start, stop))
        except KeyError:
            raise error.BASICError(error.PERMISSION_DENIED)
